import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Type

from optimagic import deprecations
from optimagic.algorithms import ALL_ALGORITHMS
from optimagic.deprecations import (
    handle_log_options_throw_deprecated_warning,
    replace_and_warn_about_deprecated_algo_options,
    replace_and_warn_about_deprecated_bounds,
)
from optimagic.differentiation.numdiff_options import (
    NumdiffOptions,
    NumdiffPurpose,
    get_default_numdiff_options,
    pre_process_numdiff_options,
)
from optimagic.exceptions import (
    AliasError,
    InvalidFunctionError,
    MissingInputError,
)
from optimagic.logging.logger import LogOptions, SQLiteLogOptions
from optimagic.optimization.algorithm import AlgoInfo, Algorithm
from optimagic.optimization.fun_value import (
    SpecificFunctionValue,
    convert_fun_output_to_function_value,
    enforce_return_type,
    enforce_return_type_with_jac,
)
from optimagic.optimization.multistart_options import (
    MultistartOptions,
    pre_process_multistart,
)
from optimagic.optimization.scipy_aliases import (
    map_method_to_algorithm,
    split_fun_and_jac,
)
from optimagic.parameters.bounds import Bounds, pre_process_bounds
from optimagic.parameters.scaling import ScalingOptions, pre_process_scaling
from optimagic.shared.process_user_function import (
    get_kwargs_from_args,
    infer_aggregation_level,
    partial_func_of_params,
)
from optimagic.typing import AggregationLevel, Direction, ErrorHandling, PyTree
from optimagic.utilities import propose_alternatives


@dataclass(frozen=True)
class OptimizationProblem:
    """Collect everything that defines the optimization problem.

    The attributes are very close to the arguments of `maximize` and `minimize` but they
    are converted to stricter types. For example, the bounds argument that can be a
    sequence of tuples, a scipy.optimize.Bounds object or an optimagic.Bounds when
    calling `maximize` or `minimize` is converted to an optimagic.Bounds object.

    All deprecated arguments are removed and all scipy aliases are replaced by their
    optimagic counterparts.

    All user provided functions are partialled if corresponding `kwargs` dictionaries
    were provided.

    # TODO: Document attributes after other todos are resolved.

    """

    fun: Callable[[PyTree], SpecificFunctionValue]
    params: PyTree
    algorithm: Algorithm
    bounds: Bounds | None
    # TODO: Only allow list[Constraint] or Constraint
    constraints: list[dict[str, Any]]
    jac: Callable[[PyTree], PyTree] | None
    fun_and_jac: Callable[[PyTree], tuple[SpecificFunctionValue, PyTree]] | None
    numdiff_options: NumdiffOptions
    # TODO: logging will become None | Logger and log_options will be removed
    error_handling: ErrorHandling
    logging: LogOptions | None
    error_penalty: dict[str, Any] | None
    scaling: ScalingOptions | None
    multistart: MultistartOptions | None
    collect_history: bool
    skip_checks: bool
    direction: Direction
    fun_eval: SpecificFunctionValue


def create_optimization_problem(
    direction,
    fun,
    params,
    algorithm,
    *,
    bounds,
    fun_kwargs,
    constraints,
    algo_options,
    jac,
    jac_kwargs,
    fun_and_jac,
    fun_and_jac_kwargs,
    numdiff_options,
    logging,
    error_handling,
    error_penalty,
    scaling,
    multistart,
    collect_history,
    skip_checks,
    # scipy aliases
    x0,
    method,
    args,
    # scipy arguments that are not yet supported
    hess,
    hessp,
    callback,
    # scipy arguments that will never be supported
    options,
    tol,
    # deprecated arguments
    criterion,
    criterion_kwargs,
    derivative,
    derivative_kwargs,
    criterion_and_derivative,
    criterion_and_derivative_kwargs,
    lower_bounds,
    log_options,
    upper_bounds,
    soft_lower_bounds,
    soft_upper_bounds,
    scaling_options,
    multistart_options,
):
    # ==================================================================================
    # error handling needed as long as fun is an optional argument
    # ==================================================================================

    if fun is None and criterion is None:
        msg = (
            "Missing objective function. Please provide an objective function as the "
            "first positional argument or as the keyword argument `fun`."
        )
        raise MissingInputError(msg)

    if params is None and x0 is None:
        msg = (
            "Missing start parameters. Please provide start parameters as the second "
            "positional argument or as the keyword argument `params`."
        )
        raise MissingInputError(msg)

    if algorithm is None and method is None:
        msg = (
            "Missing algorithm. Please provide an algorithm as the third positional "
            "argument or as the keyword argument `algorithm`."
        )
        raise MissingInputError(msg)

    # ==================================================================================
    # deprecations
    # ==================================================================================

    if log_options is not None:
        logging = handle_log_options_throw_deprecated_warning(log_options, logging)

    if criterion is not None:
        deprecations.throw_criterion_future_warning()
        fun = criterion if fun is None else fun

    if criterion_kwargs is not None:
        deprecations.throw_criterion_kwargs_future_warning()
        fun_kwargs = criterion_kwargs if fun_kwargs is None else fun_kwargs

    if derivative is not None:
        deprecations.throw_derivative_future_warning()
        jac = derivative if jac is None else jac

    if derivative_kwargs is not None:
        deprecations.throw_derivative_kwargs_future_warning()
        jac_kwargs = derivative_kwargs if jac_kwargs is None else jac_kwargs

    if criterion_and_derivative is not None:
        deprecations.throw_criterion_and_derivative_future_warning()
        fun_and_jac = criterion_and_derivative if fun_and_jac is None else fun_and_jac

    if criterion_and_derivative_kwargs is not None:
        deprecations.throw_criterion_and_derivative_kwargs_future_warning()
        fun_and_jac_kwargs = (
            criterion_and_derivative_kwargs
            if fun_and_jac_kwargs is None
            else fun_and_jac_kwargs
        )

    if scaling_options is not None:
        deprecations.throw_scaling_options_future_warning()
        if scaling is True and scaling_options is not None:
            scaling = scaling_options

    if multistart_options is not None:
        deprecations.throw_multistart_options_future_warning()
        if multistart is True and multistart_options is not None:
            multistart = multistart_options

    deprecations.throw_dict_constraints_future_warning_if_required(constraints)

    algo_options = replace_and_warn_about_deprecated_algo_options(algo_options)

    bounds = replace_and_warn_about_deprecated_bounds(
        lower_bounds=lower_bounds,
        upper_bounds=upper_bounds,
        bounds=bounds,
        soft_lower_bounds=soft_lower_bounds,
        soft_upper_bounds=soft_upper_bounds,
    )

    if isinstance(jac, dict):
        jac = deprecations.replace_and_warn_about_deprecated_derivatives(jac, "jac")

    if isinstance(fun_and_jac, dict):
        fun_and_jac = deprecations.replace_and_warn_about_deprecated_derivatives(
            fun_and_jac, "fun_and_jac"
        )
    # ==================================================================================
    # handle scipy aliases
    # ==================================================================================

    if x0 is not None:
        if params is not None:
            msg = (
                "x0 is an alias for params (for better compatibility with scipy). "
                "Do not use both x0 and params."
            )
            raise AliasError(msg)
        else:
            params = x0

    if method is not None:
        if algorithm is not None:
            msg = (
                "method is an alias for algorithm to select the scipy optimizers under "
                "their original name. Do not use both method and algorithm."
            )
            raise AliasError(msg)
        else:
            algorithm = map_method_to_algorithm(method)

    if args is not None:
        if (
            fun_kwargs is not None
            or jac_kwargs is not None
            or fun_and_jac_kwargs is not None
        ):
            msg = (
                "args is an alternative to fun_kwargs, jac_kwargs and "
                "fun_and_jac_kwargs that optimagic supports for compatibility "
                "with scipy. Do not use args in conjunction with any of the other "
                "arguments."
            )
            raise AliasError(msg)
        else:
            kwargs = get_kwargs_from_args(args, fun, offset=1)
            fun_kwargs, jac_kwargs, fun_and_jac_kwargs = kwargs, kwargs, kwargs

    # jac is not an alias but we need to handle the case where `jac=True`, i.e. fun is
    # actually fun_and_jac. This is not recommended in optimagic because then optimizers
    # cannot evaluate fun in isolation but we can easily support it for compatibility.
    if jac is True:
        jac = None
        if fun_and_jac is None:
            fun_and_jac = fun
            fun = split_fun_and_jac(fun_and_jac, target="fun")

    # ==================================================================================
    # Handle scipy arguments that are not yet implemented
    # ==================================================================================

    if hess is not None:
        msg = (
            "The hess argument is not yet supported in optimagic. Creat an issue on "
            "https://github.com/optimagic-dev/optimagic/ if you have urgent need "
            "for this feature."
        )
        raise NotImplementedError(msg)

    if hessp is not None:
        msg = (
            "The hessp argument is not yet supported in optimagic. Creat an issue on "
            "https://github.com/optimagic-dev/optimagic/ if you have urgent need "
            "for this feature."
        )
        raise NotImplementedError(msg)

    if callback is not None:
        msg = (
            "The callback argument is not yet supported in optimagic. Creat an issue "
            "on https://github.com/optimagic-dev/optimagic/ if you have urgent "
            "need for this feature."
        )
        raise NotImplementedError(msg)

    # ==================================================================================
    # Handle scipy arguments that will never be supported
    # ==================================================================================

    if options is not None:
        # TODO: Add link to a how-to guide or tutorial for this
        msg = (
            "The options argument is not supported in optimagic. Please use the "
            "algo_options argument instead."
        )
        raise NotImplementedError(msg)

    if tol is not None:
        # TODO: Add link to a how-to guide or tutorial for this
        msg = (
            "The tol argument is not supported in optimagic. Please use "
            "algo_options or configured algorithms instead to set convergence criteria "
            "for your optimizer."
        )
        raise NotImplementedError(msg)

    # ==================================================================================
    # Convert literals to enums
    # ==================================================================================
    error_handling = ErrorHandling(error_handling)

    # ==================================================================================
    # Set default values and check options
    # ==================================================================================
    bounds = pre_process_bounds(bounds)
    scaling = pre_process_scaling(scaling)
    multistart = pre_process_multistart(multistart)
    numdiff_options = pre_process_numdiff_options(numdiff_options)
    constraints = deprecations.pre_process_constraints(constraints)

    if numdiff_options is None:
        numdiff_options = get_default_numdiff_options(purpose=NumdiffPurpose.OPTIMIZE)

    fun_kwargs = {} if fun_kwargs is None else fun_kwargs
    constraints = [] if constraints is None else constraints
    algo_options = {} if algo_options is None else algo_options
    jac_kwargs = {} if jac_kwargs is None else jac_kwargs
    fun_and_jac_kwargs = {} if fun_and_jac_kwargs is None else fun_and_jac_kwargs
    error_penalty = {} if error_penalty is None else error_penalty

    if isinstance(logging, str) or isinstance(logging, Path):
        log_path = Path(logging)
        logging = SQLiteLogOptions(log_path)

    # ==================================================================================
    # evaluate fun for the first time
    # ==================================================================================
    fun = partial_func_of_params(
        func=fun,
        kwargs=fun_kwargs,
        name="criterion",
        skip_checks=skip_checks,
    )

    # This should be done as late as possible; It has to be done here to infer the
    # problem type until the decorator approach becomes mandatory.
    # TODO: Move this into `_optimize` as soon as we reach 0.6.0
    try:
        fun_eval = fun(params)
    except (KeyboardInterrupt, SystemExit):
        raise
    except Exception as e:
        msg = "Error while evaluating fun at start params."
        raise InvalidFunctionError(msg) from e

    if deprecations.is_dict_output(fun_eval):
        deprecations.throw_dict_output_warning()

    # ==================================================================================
    # infer the problem type
    # ==================================================================================

    if deprecations.is_dict_output(fun_eval):
        problem_type = deprecations.infer_problem_type_from_dict_output(fun_eval)
    else:
        problem_type = infer_aggregation_level(fun)

    if (
        problem_type == AggregationLevel.LEAST_SQUARES
        and direction == Direction.MAXIMIZE
    ):
        raise InvalidFunctionError("Least-squares problems cannot be maximized.")

    # ==================================================================================
    # process the fun_eval; Can be removed once the first evaluation gets moved to
    # a later point where the `enforce` decorator has already been applied.
    # ==================================================================================
    if deprecations.is_dict_output(fun_eval):
        fun_eval = deprecations.convert_dict_to_function_value(fun_eval)
        fun = deprecations.replace_dict_output(fun)
    else:
        fun_eval = convert_fun_output_to_function_value(fun_eval, problem_type)

    fun = enforce_return_type(problem_type)(fun)

    # ==================================================================================
    # Process the user provided algorithm
    # ==================================================================================

    algorithm = pre_process_user_algorithm(algorithm)
    algorithm = algorithm.with_option_if_applicable(**algo_options)

    if algorithm.algo_info.solver_type == AggregationLevel.LIKELIHOOD:
        if problem_type not in [
            AggregationLevel.LIKELIHOOD,
            AggregationLevel.LEAST_SQUARES,
        ]:
            raise InvalidFunctionError(
                "Likelihood solvers can only be used with likelihood or least-squares "
                "problems."
            )
    elif algorithm.algo_info.solver_type == AggregationLevel.LEAST_SQUARES:
        if problem_type != AggregationLevel.LEAST_SQUARES:
            raise InvalidFunctionError(
                "Least-squares solvers can only be used with least-squares problems."
            )

    # ==================================================================================
    # select the correct derivative functions
    # ==================================================================================

    if jac is not None:
        jac = pre_process_derivatives(
            candidate=jac, name="jac", solver_type=algorithm.algo_info.solver_type
        )

    if fun_and_jac is not None:
        fun_and_jac = pre_process_derivatives(
            candidate=fun_and_jac,
            name="fun_and_jac",
            solver_type=algorithm.algo_info.solver_type,
        )

    # ==================================================================================
    # partial the kwargs into corresponding functions
    # ==================================================================================

    if jac is not None:
        jac = partial_func_of_params(
            func=jac,
            kwargs=jac_kwargs,
            name="derivative",
            skip_checks=skip_checks,
        )

    if fun_and_jac is not None:
        fun_and_jac = partial_func_of_params(
            func=fun_and_jac,
            kwargs=fun_and_jac_kwargs,
            name="criterion_and_derivative",
            skip_checks=skip_checks,
        )
        fun_and_jac = deprecations.replace_dict_output(fun_and_jac)

        fun_and_jac = enforce_return_type_with_jac(algorithm.algo_info.solver_type)(
            fun_and_jac
        )

    # ==================================================================================
    # Check types of arguments
    # ==================================================================================

    if not skip_checks:
        if params is None:
            raise ValueError("params cannot be None")

        if not isinstance(fun, Callable):
            raise ValueError("fun must be a callable")

        if not isinstance(algorithm, Algorithm):
            raise ValueError("algorithm must be an Algorithm object.")

        if not isinstance(algo_options, dict | None):
            raise ValueError("algo_options must be a dictionary or None")

        if not isinstance(algorithm.algo_info, AlgoInfo):
            raise ValueError("algo_info must be an AlgoInfo object")

        if not isinstance(bounds, Bounds | None):
            raise ValueError("bounds must be a Bounds object or None")

        if not all(isinstance(c, dict) for c in constraints):
            # TODO: Only allow list[Constraint]
            raise ValueError("constraints must be a list of dictionaries")

        if not isinstance(jac, Callable | None):
            raise ValueError("jac must be a callable or None")

        if not isinstance(fun_and_jac, Callable | None):
            raise ValueError("fun_and_jac must be a callable or None")

        if not isinstance(numdiff_options, NumdiffOptions):
            raise ValueError("numdiff_options must be a NumdiffOptions object")

        if not isinstance(logging, bool | Path | LogOptions | None):
            raise ValueError(
                "logging must be a boolean, a path, a LogOptions instance or None"
            )

        if not isinstance(log_options, dict | None):
            raise ValueError("log_options must be a dictionary or None")

        if not isinstance(error_penalty, dict | None):
            raise ValueError("error_penalty must be a dictionary or None")

        if not isinstance(scaling, ScalingOptions | None):
            raise ValueError("scaling must be a ScalingOptions object or None")

        if not isinstance(multistart, MultistartOptions | None):
            raise ValueError("multistart must be a MultistartOptions object or None")

        if not isinstance(collect_history, bool):
            raise ValueError("collect_history must be a boolean")

    # ==================================================================================
    # create the problem object
    # ==================================================================================

    problem = OptimizationProblem(
        fun=fun,
        params=params,
        algorithm=algorithm,
        bounds=bounds,
        constraints=constraints,
        jac=jac,
        fun_and_jac=fun_and_jac,
        numdiff_options=numdiff_options,
        logging=logging,
        error_handling=error_handling,
        error_penalty=error_penalty,
        scaling=scaling,
        multistart=multistart,
        collect_history=collect_history,
        skip_checks=skip_checks,
        direction=direction,
        fun_eval=fun_eval,
    )

    return problem


def pre_process_derivatives(candidate, name, solver_type):
    if callable(candidate):
        candidate = [candidate]

    out = None
    for func in candidate:
        if not callable(func):
            raise ValueError(f"{name} must be a callable or sequence of callables.")

        problem_type = infer_aggregation_level(func)
        if problem_type == solver_type:
            out = func

    if out is None:
        msg = (
            f"You used the `{name}` argument but none of the callables you provided "
            "has the correct aggregation level for your selected optimization "
            "algorithm. Falling back to numerical derivatives."
        )
        warnings.warn(msg)

    return out


def pre_process_user_algorithm(
    algorithm: str | Algorithm | Type[Algorithm],
) -> Algorithm:
    """Process the user specfied algorithm."""
    if isinstance(algorithm, str):
        try:
            # Use ALL_ALGORITHMS and not just AVAILABLE_ALGORITHMS such that the
            # algorithm specific error message with installation instruction will be
            # reached if an optional dependency is not installed.
            algorithm = ALL_ALGORITHMS[algorithm]()
        except KeyError:
            proposed = propose_alternatives(algorithm, list(ALL_ALGORITHMS))
            raise ValueError(
                f"Invalid algorithm: {algorithm}. Did you mean {proposed}?"
            ) from None
    elif isinstance(algorithm, type) and issubclass(algorithm, Algorithm):
        algorithm = algorithm()

    return algorithm



from typing import Any

from optimagic.typing import (
    GtOneFloat,
    NonNegativeFloat,
    NonNegativeInt,
    PositiveFloat,
    PositiveInt,
)


def _process_float_like(value: Any) -> float:
    """Process a value that should be converted to a float."""
    return float(value)


def _process_int_like(value: Any) -> int:
    """Process a value that should be converted to an int."""
    if isinstance(value, int):
        return value
    elif isinstance(value, str):
        return int(float(value))
    else:
        return int(value)


def _process_positive_int_like(value: Any) -> PositiveInt:
    """Process a value that should be converted to a positive int."""
    out = _process_int_like(value)
    if out <= 0:
        raise ValueError(f"Value must be positive, got {out}")
    return out


def _process_non_negative_int_like(value: Any) -> NonNegativeInt:
    """Process a value that should be converted to a non-negative int."""
    out = _process_int_like(value)
    if out < 0:
        raise ValueError(f"Value must be non-negative, got {out}")
    return out


def _process_positive_float_like(value: Any) -> PositiveFloat:
    """Process a value that should be converted to a positive float."""
    out = _process_float_like(value)
    if out <= 0:
        raise ValueError(f"Value must be positive, got {out}")
    return out


def _process_non_negative_float_like(value: Any) -> NonNegativeFloat:
    """Process a value that should be converted to a non-negative float."""
    out = _process_float_like(value)
    if out < 0:
        raise ValueError(f"Value must be non-negative, got {out}")
    return out


def _process_gt_one_float_like(value: Any) -> GtOneFloat:
    """Process a value that should be converted to a float greater than one."""
    out = _process_float_like(value)
    if out <= 1:
        raise ValueError(f"Value must be greater than one, got {out}")
    return out


def _process_bool_like(value: Any) -> bool:
    """Process a value that should be converted to a bool."""
    if isinstance(value, bool):
        return value
    elif isinstance(value, str):
        if value.lower() in {"true", "1", "yes"}:
            return True
        elif value.lower() in {"false", "0", "no"}:
            return False

    return bool(value)


TYPE_CONVERTERS = {
    float: _process_float_like,
    int: _process_int_like,
    bool: _process_bool_like,
    PositiveInt: _process_positive_int_like,
    NonNegativeInt: _process_non_negative_int_like,
    PositiveFloat: _process_positive_float_like,
    NonNegativeFloat: _process_non_negative_float_like,
    GtOneFloat: _process_gt_one_float_like,
}



from dataclasses import dataclass
from enum import Enum
from typing import Literal

from optimagic.optimization.fun_value import SpecificFunctionValue
from optimagic.typing import (
    DictLikeAccess,
    Direction,
    DirectionLiteral,
    PyTree,
)


class StepStatus(str, Enum):
    """Status of a step in a process.

    Attributes:
        SCHEDULED: Indicates that the step is scheduled but not yet started.
        RUNNING: Indicates that the step is currently in progress.
        COMPLETE: Indicates that the step has completed successfully.
        SKIPPED: Indicates that the step was skipped.

    """

    SCHEDULED = "scheduled"
    RUNNING = "running"
    COMPLETE = "complete"
    SKIPPED = "skipped"


StepStatusLiteral = Literal["scheduled", "running", "complete", "skipped"]


class StepType(str, Enum):
    """Type of step in a process.

    Attributes:
        OPTIMIZATION: Represents an optimization step.
        EXPLORATION: Represents an exploration step.

    """

    OPTIMIZATION = "optimization"
    EXPLORATION = "exploration"


StepTypeLiteral = Literal["optimization", "exploration"]


class ExistenceStrategy(str, Enum):
    """Strategies to handle the existence of a database or table.

    Attributes:
        RAISE: Raise an error if the database resp. table exists.
        EXTEND: Extend the existing database or table.
        REPLACE: Replace the existing database or table.

    """

    RAISE = "raise"
    EXTEND = "extend"
    REPLACE = "replace"


ExistenceStrategyLiteral = Literal["raise", "extend", "replace"]


@dataclass(frozen=True)
class IterationState(DictLikeAccess):
    """Result of a criterion evaluation.

    Attributes:
        params: The parameters used in the evaluation.
        timestamp: The time at which the evaluation was performed.
        value: The result value of the evaluation.
        valid: Indicates if the evaluation is valid.
        criterion_eval: Optional, additional evaluation information.
        internal_derivative: Optional, derivative information used internally.
        step: Optional, step number associated with the evaluation.
        exceptions: Optional, exceptions encountered during evaluation.
        hash: Optional, hash of the evaluation for identification purposes.

    """

    params: PyTree
    timestamp: float
    scalar_fun: float | None
    valid: bool
    raw_fun: SpecificFunctionValue | None
    step: int | None
    exceptions: str | None

    def combine(self, other: "IterationState") -> "IterationState":
        """Combine two iteration states.

        Args:
            other (IterationState): The second iteration state.

        Returns:
            IterationState: The combined iteration state.

        """
        raw = [e for e in [self.exceptions, other.exceptions] if e is not None]
        exceptions: str | None = None
        if raw:
            exceptions = "\n\n".join(raw)

        new = IterationState(
            # one of the values must be None
            params=self.params,
            timestamp=min(self.timestamp, other.timestamp),
            scalar_fun=self.scalar_fun or other.scalar_fun,
            valid=self.valid and other.valid,
            # one of the values must be None
            raw_fun=self.raw_fun or other.raw_fun,
            step=self.step,
            exceptions=exceptions,
        )
        return new


@dataclass(frozen=True)
class IterationStateWithId(IterationState):
    """Criterion evaluation result with an ID.

    Attributes:
        rowid: The unique ID associated with the evaluation result.

    Raises:
        ValueError: If `rowid` is None.

    """

    rowid: int | None = None

    def __post_init__(self) -> None:
        if self.rowid is None:
            raise ValueError("rowid must not be None")


@dataclass(frozen=True)
class StepResult(DictLikeAccess):
    """Result of a process step.

    Attributes:
        name: The name of the step.
        type: The type of the step, either as `StepType` or string.
        status: The status of the step, either as `StepStatus` or string.
        n_iterations: Optional, the number of iterations performed in the step.

    """

    name: str
    type: StepType | StepTypeLiteral
    status: StepStatus | StepStatusLiteral
    n_iterations: int | None = None

    def __post_init__(self) -> None:
        if isinstance(self.type, str):
            object.__setattr__(self, "type", StepType(self.type))
        if isinstance(self.status, str):
            object.__setattr__(self, "status", StepStatus(self.status))


@dataclass(frozen=True)
class StepResultWithId(StepResult):
    """Step result with an ID.

    Attributes:
        rowid: The unique ID associated with the step result.

    Raises:
        ValueError: If `rowid` is None.

    """

    rowid: int | None = None

    def __post_init__(self) -> None:
        if self.rowid is None:
            raise ValueError("rowid must not be None")
        super().__post_init__()


@dataclass(frozen=True)
class ProblemInitialization(DictLikeAccess):
    """Start characteristics of an optimization problem.

    Attributes:
        direction: The direction of optimization,
            either as `Direction` or string literal.
        params: The parameters for the initialization.

    """

    direction: Direction | DirectionLiteral
    params: PyTree


@dataclass(frozen=True)
class ProblemInitializationWithId(ProblemInitialization):
    """Problem initialization with an ID.

    Attributes:
        rowid: The unique ID associated with the problem initialization.

    Raises:
        ValueError: If `rowid` is None.

    """

    rowid: int | None = None

    def __post_init__(self) -> None:
        if self.rowid is None:
            raise ValueError("rowid must not be None")



from typing import Any, cast

from optimagic.logging.logger import LogStore
from optimagic.logging.types import StepResult, StepStatus


def log_scheduled_steps_and_get_ids(
    steps: list[dict[str, Any]], logger: LogStore | None
) -> list[int]:
    """Add scheduled steps to the steps table of the database and get their ids.

    The ids are only determined once the steps are written to the database and the
    ids of all previously existing steps are known.

    Args:
        steps (list): List of dicts with entries for the steps table.
        logging (bool): Whether to actually write to the database.

    Returns:
        list: List of integers with the step ids.

    """
    default_row = {"status": StepStatus.SCHEDULED.value}
    if logger:
        for row in steps:
            data = StepResult(**{**default_row, **row})
            logger.step_store.insert(data)

        last_steps = logger.step_store.select_last_rows(len(steps))
        step_ids = cast(list[int], [row.rowid for row in last_steps])
    else:
        step_ids = list(range(len(steps)))

    return step_ids



from __future__ import annotations

import os
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Generic, Type, TypeVar, cast

import numpy as np
import pandas as pd
import sqlalchemy as sql
from sqlalchemy.engine import Engine

from optimagic.logging.base import (
    NonUpdatableKeyValueStore,
    UpdatableKeyValueStore,
)
from optimagic.logging.sqlalchemy import (
    IterationStore,
    ProblemStore,
    SQLAlchemyConfig,
    StepStore,
)
from optimagic.logging.types import (
    ExistenceStrategy,
    ExistenceStrategyLiteral,
    IterationState,
    IterationStateWithId,
    ProblemInitialization,
    ProblemInitializationWithId,
    StepResult,
    StepResultWithId,
    StepType,
)
from optimagic.typing import (
    Direction,
    DirectionLiteral,
    IterationHistory,
    MultiStartIterationHistory,
    PyTree,
)


class LogOptions:
    """Base class for defining different log options.

    Serves as a registry for implemented option classes for better discoverability.

    """

    _subclass_registry: list[Type[LogOptions]] = []

    def __init_subclass__(
        cls: Type[LogOptions], abstract: bool = False, **kwargs: dict[Any, Any]
    ):
        if not abstract:
            LogOptions._subclass_registry.append(cls)
        super().__init_subclass__(**kwargs)

    @classmethod
    def available_option_types(cls) -> list[Type[LogOptions]]:
        return cls._subclass_registry


_LogOptionsType = TypeVar("_LogOptionsType", bound=LogOptions)


class LogReader(Generic[_LogOptionsType], ABC):
    """A class that manages the retrieving of optimization and exploration data.

    This class exposes methods to retrieve optimization logging data from stores.

    """

    _step_store: UpdatableKeyValueStore[StepResult, StepResultWithId]
    _iteration_store: NonUpdatableKeyValueStore[IterationState, IterationStateWithId]
    _problem_store: UpdatableKeyValueStore[
        ProblemInitialization, ProblemInitializationWithId
    ]

    @property
    def problem_df(self) -> pd.DataFrame:
        return self._problem_store.to_df()

    @classmethod
    def from_options(cls, log_options: LogOptions) -> LogReader[_LogOptionsType]:
        log_reader_class = _LOG_OPTION_LOG_READER_REGISTRY.get(type(log_options), None)

        if log_reader_class is None:
            raise ValueError(
                f"No LogReader implementation found for type "
                f"{type(log_options)}. Available option types: "
                f"\n {list(_LOG_OPTION_LOG_READER_REGISTRY.keys())}"
            )

        return log_reader_class._create(log_options)

    @classmethod
    @abstractmethod
    def _create(cls, log_options: _LogOptionsType) -> LogReader[_LogOptionsType]:
        pass

    def read_iteration(self, iteration: int) -> IterationStateWithId:
        """Read a specific iteration from the iteration store.

        Args:
            iteration: The iteration number to read. Negative values read from the end.

        Returns:
            A `CriterionEvaluationWithId` object containing the iteration data.

        Raises:
            IndexError: If the iteration is invalid or the store is empty.

        """
        if iteration >= 0:
            rowid = iteration + 1
        else:
            try:
                last_row = self._iteration_store.select_last_rows(1)
                highest_rowid = last_row[0].rowid
            except IndexError as e:
                raise IndexError(
                    "Invalid iteration request, iteration store is empty"
                ) from e

            # iteration is negative here!
            assert highest_rowid is not None
            rowid = highest_rowid + iteration + 1

        row_list = self._iteration_store.select(rowid)

        if len(row_list) == 0:
            raise IndexError(f"Invalid iteration requested: {iteration}")
        else:
            data = row_list[0]

        return data

    def read_history(self) -> IterationHistory:
        """Read the entire iteration history from the iteration store.

        Returns:
            An `IterationHistory` object containing the parameters,
                criterion values, and runtimes.

        """
        raw_res = self._iteration_store.select()
        params_list = []
        criterion_list = []
        runtime_list = []
        for data in raw_res:
            if data.scalar_fun is not None:
                params_list.append(data.params)
                criterion_list.append(data.scalar_fun)
                runtime_list.append(data.timestamp)

        times = np.array(runtime_list)
        times -= times[0]

        return IterationHistory(params_list, criterion_list, times)

    @staticmethod
    def _normalize_direction(
        direction: Direction | DirectionLiteral,
    ) -> Direction:
        if isinstance(direction, str):
            direction = Direction(direction)
        return direction

    def _build_history_dataframe(self) -> pd.DataFrame:
        steps = self._step_store.to_df()
        raw_res = self._iteration_store.select()

        history: dict[str, list[Any]] = {
            "params": [],
            "fun": [],
            "time": [],
            "step": [],
        }

        for data in raw_res:
            if data.scalar_fun is not None:
                history["params"].append(data.params)
                history["fun"].append(data.scalar_fun)
                history["time"].append(data.timestamp)
                history["step"].append(data.step)

        times = np.array(history["time"])
        times -= times[0]
        history["time"] = times.tolist()

        df = pd.DataFrame(history)
        df = df.merge(
            steps[[f"{self._step_store.primary_key}", "type"]],
            left_on="step",
            right_on=f"{self._step_store.primary_key}",
        )
        return df.drop(columns=f"{self._step_store.primary_key}")

    @staticmethod
    def _split_exploration_and_optimization(
        df: pd.DataFrame,
    ) -> tuple[pd.DataFrame | None, pd.DataFrame]:
        exploration = df.query(f"type == '{StepType.EXPLORATION.value}'").drop(
            columns=["step", "type"]
        )
        histories = df.query(f"type == '{StepType.OPTIMIZATION.value}'")
        histories = histories.drop(columns="type").set_index("step", append=True)

        return None if exploration.empty else exploration, histories

    @staticmethod
    def _sort_exploration(
        exploration: pd.DataFrame | None, optimization_type: Direction
    ) -> IterationHistory | None:
        if exploration is not None:
            is_minimization = optimization_type is Direction.MINIMIZE
            exploration = exploration.sort_values(by="fun", ascending=is_minimization)
            exploration_dict = cast(dict[str, Any], exploration.to_dict(orient="list"))
            return IterationHistory(**exploration_dict)
        return exploration

    @staticmethod
    def _extract_best_history(
        histories: pd.DataFrame, optimization_type: Direction
    ) -> tuple[IterationHistory, list[IterationHistory] | None]:
        groupby_step_criterion = histories["fun"].groupby(level="step")

        if optimization_type is Direction.MINIMIZE:
            best_idx = groupby_step_criterion.min().idxmin()
        else:
            best_idx = groupby_step_criterion.max().idxmax()

        remaining_indices = (
            histories.index.get_level_values("step").unique().difference([best_idx])
        )

        best_history: pd.DataFrame | pd.Series[Any] = histories.xs(
            best_idx, level="step"
        )

        def _to_dict(pandas_obj: pd.DataFrame | pd.Series) -> dict[str, Any]:  # type:ignore
            if isinstance(pandas_obj, pd.DataFrame):
                result = pandas_obj.to_dict(orient="list")
            else:
                result = best_history.to_dict()
            return cast(dict[str, Any], result)

        best_history_dict = _to_dict(best_history)
        local_histories = [
            _to_dict(histories.xs(idx, level="step")) for idx in remaining_indices
        ]
        if len(local_histories) == 0:
            remaining_histories = None
        else:
            remaining_histories = [
                IterationHistory(**history_dict) for history_dict in local_histories
            ]

        return IterationHistory(**best_history_dict), remaining_histories

    def read_multistart_history(
        self, direction: Direction | DirectionLiteral
    ) -> MultiStartIterationHistory:
        """Read and the multistart optimization history.

        Args:
            direction: The optimization direction, either as an enum or string.

        Returns:
            A `MultiStartIterationHistory` object containing the best history,
                local histories, and exploration history.

        """
        optimization_type = self._normalize_direction(direction)
        history_df = self._build_history_dataframe()
        exploration, optimization_history = self._split_exploration_and_optimization(
            history_df
        )
        exploration_history = self._sort_exploration(exploration, optimization_type)
        best_history, remaining_histories = self._extract_best_history(
            optimization_history, optimization_type
        )

        return MultiStartIterationHistory(
            best_history,
            local_histories=remaining_histories,
            exploration=exploration_history,
        )

    def read_start_params(self) -> PyTree:
        """Read the start parameters form the problem store.

        Returns:
            A pytree object representing the start parameter.

        """
        return self._problem_store.select(1)[0].params


_LogReaderType = TypeVar("_LogReaderType", bound=LogReader[Any])


class LogStore(Generic[_LogOptionsType, _LogReaderType], ABC):
    """A class that manages the logging of optimization and exploration data.

    This class handles storing iterations, steps, and problem
    initialization data using various stores.

    Args:
        iteration_store: A non-updatable store for iteration data.
        step_store: An updatable store for step data.
        problem_store: An updatable store for problem initialization data.

    """

    def __init__(
        self,
        iteration_store: NonUpdatableKeyValueStore[
            IterationState, IterationStateWithId
        ],
        step_store: UpdatableKeyValueStore[StepResult, StepResultWithId],
        problem_store: UpdatableKeyValueStore[
            ProblemInitialization, ProblemInitializationWithId
        ],
    ):
        self.step_store = step_store
        self.iteration_store = iteration_store
        self.problem_store = problem_store

    @classmethod
    def from_options(
        cls, log_options: LogOptions
    ) -> LogStore[_LogOptionsType, _LogReaderType]:
        logger_class = _LOG_OPTION_LOGGER_REGISTRY.get(type(log_options), None)

        if logger_class is None:
            raise ValueError(
                f"No Logger implementation found for type "
                f"{type(log_options)}. Available option types: "
                f"\n {list(_LOG_OPTION_LOGGER_REGISTRY.keys())}"
            )

        return logger_class.create(log_options)

    @classmethod
    @abstractmethod
    def create(
        cls, log_options: _LogOptionsType
    ) -> LogStore[_LogOptionsType, _LogReaderType]:
        pass


class SQLiteLogOptions(SQLAlchemyConfig, LogOptions):
    """Configuration class for setting up an SQLite database with SQLAlchemy.

    This class extends the `SQLAlchemyConfig` class to configure an SQLite database.
    It handles the creation of the database engine, manages database files,
    and applies various optimizations for logging performance.

    Args:
        path (str | Path): The file path to the SQLite database.
        fast_logging (bool): A boolean that determines if “unsafe” settings are used to
            speed up write processes to the database. This should only be used for very
            short running criterion functions where the main purpose of the log
            is a real-time dashboard, and it would not be catastrophic to get
            a corrupted database in case of a sudden system shutdown.
            If one evaluation of the criterion function (and gradient if applicable)
            takes more than 100 ms, the logging overhead is negligible.
        if_database_exists (ExistenceStrategy): Strategy for handling an existing
            database file. One of “extend”, “replace”, “raise”.

    """

    def __init__(
        self,
        path: str | Path,
        fast_logging: bool = True,
        if_database_exists: ExistenceStrategy
        | ExistenceStrategyLiteral = ExistenceStrategy.RAISE,
    ):
        url = f"sqlite:///{path}"
        self._fast_logging = fast_logging
        self._path = path
        if isinstance(if_database_exists, str):
            if_database_exists = ExistenceStrategy(if_database_exists)
        self.if_database_exists = if_database_exists
        super().__init__(url)

    @property
    def path(self) -> str | Path:
        return self._path

    def create_engine(self) -> Engine:
        engine = sql.create_engine(self.url)
        self._configure_engine(engine)
        return engine

    def _configure_engine(self, engine: Engine) -> None:
        """Configure the sqlite engine.

        The two functions that configure the emission of the `begin` statement are taken
        from the sqlalchemy documentation the documentation:
        https://tinyurl.com/u9xea5z
        and are
        the recommended way of working around a bug in the pysqlite driver.

        The other function speeds up the write process. If fast_logging is False, it
        does so using only completely safe optimizations. Of fast_logging is True,
        it also uses unsafe optimizations.

        """

        @sql.event.listens_for(engine, "connect")
        def do_connect(dbapi_connection: Any, connection_record: Any) -> None:  # noqa: ARG001
            # disable pysqlite's emitting of the BEGIN statement entirely.
            # also stops it from emitting COMMIT before absolutely necessary.
            dbapi_connection.isolation_level = None

        @sql.event.listens_for(engine, "begin")
        def do_begin(conn: Any) -> None:
            # emit our own BEGIN
            conn.exec_driver_sql("BEGIN DEFERRED")

        @sql.event.listens_for(engine, "connect")
        def set_sqlite_pragma(dbapi_connection: Any, connection_record: Any) -> None:  # noqa: ARG001
            cursor = dbapi_connection.cursor()
            cursor.execute("PRAGMA journal_mode = WAL")
            if self._fast_logging:
                cursor.execute("PRAGMA synchronous = OFF")
            else:
                cursor.execute("PRAGMA synchronous = NORMAL")
            cursor.close()


class SQLiteLogReader(LogReader[SQLiteLogOptions]):
    """A class that manages the retrieving of optimization and exploration data from a
    SQLite database.

    This class exposes methods to retrieve optimization logging data from stores.

    Args:
            path (str | Path): The path to the SQLite database file.

    """

    def __init__(self, path: str | Path):
        if not os.path.exists(path):
            raise FileNotFoundError(f"No file found at {path=}")

        log_options = SQLiteLogOptions(
            path, fast_logging=True, if_database_exists=ExistenceStrategy.EXTEND
        )
        self._iteration_store = IterationStore(log_options)
        self._step_store = StepStore(log_options)
        self._problem_store = ProblemStore(log_options)

    @classmethod
    def _create(cls, log_options: SQLiteLogOptions) -> SQLiteLogReader:
        """Create an instance of SQLiteLogReader using the provided log options.

        Args:
            log_options (SQLiteLogOptions): Configuration options for the SQLite log.

        Returns:
            SQLiteLogReader: An instance of SQLiteLogReader initialized with the
            provided log options.

        """
        return cls(log_options.path)


class _SQLiteLogStore(LogStore[SQLiteLogOptions, SQLiteLogReader]):
    """A logger class that stores and manages optimization and exploration data using
    SQLite.

    It supports different strategies for handling existing databases, such as extending,
    replacing, or raising an error.

    """

    @staticmethod
    def _handle_existing_database(
        path: str | Path,
        if_database_exists: ExistenceStrategy | ExistenceStrategyLiteral,
    ) -> None:
        if isinstance(if_database_exists, str):
            if_database_exists = ExistenceStrategy(if_database_exists)
        database_exists = os.path.exists(path)
        if database_exists:
            if if_database_exists is ExistenceStrategy.RAISE:
                raise FileExistsError(
                    f"The database at {path} already exists. To reuse and extend "
                    f"the existing database, set if_database_exists to "
                    f"ExistenceStrategy.EXTEND."
                )
            elif if_database_exists is ExistenceStrategy.REPLACE:
                try:
                    os.remove(path)
                except PermissionError as e:
                    msg = (
                        f"Failed to remove file {path}. "
                        f"In particular, this can happen on Windows "
                        f"machines, when a different process is accessing the file, "
                        f"which results in a PermissionError. In this case, delete"
                        f"the file manually."
                    )
                    raise RuntimeError(msg) from e

    @classmethod
    def create(cls, log_options: SQLiteLogOptions) -> _SQLiteLogStore:
        cls._handle_existing_database(log_options.path, log_options.if_database_exists)

        iteration_store = IterationStore(log_options)
        step_store = StepStore(log_options)
        problem_store = ProblemStore(log_options)
        return cls(iteration_store, step_store, problem_store)


_LOG_OPTION_LOGGER_REGISTRY: dict[Type[LogOptions], Type[LogStore[Any, Any]]] = {
    SQLiteLogOptions: _SQLiteLogStore
}
_LOG_OPTION_LOG_READER_REGISTRY: dict[Type[LogOptions], Type[LogReader[Any]]] = {
    SQLiteLogOptions: SQLiteLogReader
}



"""Implement pygmo optimizers.

Notes for converting to the new algorithm interface:

- `create_algo_options` is not needed anymore because the only thing it did was mixing
  options that are supported by all optimizers (e.g. population_size, seed) with
  specific options. Then later they had to be taken appart again. Instead you need
  to pass  a few more arguments directly to `_minimize_pygmo`.
- Calling `_check_that_every_param_is_bounded` is not needed anymore. I do that check
  once in `_minimize_pygmo`.
- The documentation often just says float where I suspect PositiveFloats; Leave it at
  float for now and add todos where needed. Don't spend a lot of time on it.
- There are some specific type checks and type conversions that should not be needed
  anymore after switching to the new interface.
- Whenever we had batch_evaluator as `algo_option` we don't need it anymore but we
  should have `n_cores` in those algorithms.

"""

from __future__ import annotations

import warnings
from dataclasses import dataclass
from typing import Any, List, Literal

import numpy as np
from numpy.typing import NDArray

from optimagic import mark
from optimagic.config import IS_PYGMO_INSTALLED
from optimagic.exceptions import NotInstalledError
from optimagic.optimization.algo_options import (
    CONVERGENCE_XTOL_REL,
    STOPPING_MAXFUN_GLOBAL,
    get_population_size,
)
from optimagic.optimization.algorithm import Algorithm, InternalOptimizeResult
from optimagic.optimization.internal_optimization_problem import (
    InternalOptimizationProblem,
)
from optimagic.typing import (
    AggregationLevel,
    NonNegativeFloat,
    PositiveFloat,
    PositiveInt,
)

STOPPING_MAX_ITERATIONS_GENETIC = 250

try:
    import pygmo as pg
except ImportError:
    pass


@mark.minimizer(
    name="pygmo_gaco",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_PYGMO_INSTALLED,
    is_global=True,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=True,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class PygmoGaco(Algorithm):
    population_size: int | None = None
    n_cores: int = 1
    seed: int | None = None
    discard_start_params: bool = False
    stopping_maxiter: PositiveInt = STOPPING_MAX_ITERATIONS_GENETIC
    kernel_size: PositiveInt = 63
    speed_parameter_q: PositiveFloat = 1.0
    oracle: float = 0.0
    accuracy: PositiveFloat = 0.01
    threshold: PositiveInt = 1
    speed_of_std_values_convergence: int = 7
    stopping_max_n_without_improvements: PositiveInt = 100000
    stopping_maxfun: PositiveInt = STOPPING_MAXFUN_GLOBAL
    focus: NonNegativeFloat = 0.0
    cache: bool = False

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        population_size = get_population_size(
            population_size=self.population_size, x=x0, lower_bound=64
        )

        algo_specific_options = {
            "gen": self.stopping_maxiter,
            "ker": self.kernel_size,
            "q": self.speed_parameter_q,
            "oracle": self.oracle,
            "acc": self.accuracy,
            "threshold": self.threshold,
            "n_gen_mark": self.speed_of_std_values_convergence,
            "impstop": self.stopping_max_n_without_improvements,
            "evalstop": self.stopping_maxfun,
            "focus": self.focus,
            "memory": self.cache,
        }

        res = _minimize_pygmo(
            problem=problem,
            x0=x0,
            method="gaco",
            specific_options=algo_specific_options,
            population_size=population_size,
            n_cores=self.n_cores,
            seed=self.seed,
            discard_start_params=self.discard_start_params,
        )

        return res


@mark.minimizer(
    name="pygmo_bee_colony",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_PYGMO_INSTALLED,
    is_global=True,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class PygmoBeeColony(Algorithm):
    stopping_maxiter: PositiveInt = STOPPING_MAX_ITERATIONS_GENETIC
    seed: int | None = None
    discard_start_params: bool = False
    max_n_trials: PositiveInt = 1
    population_size: int | None = None

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        population_size = get_population_size(
            population_size=self.population_size, x=x0, lower_bound=20
        )

        algo_specific_options = {
            "limit": self.max_n_trials,
            "gen": self.stopping_maxiter,
        }

        res = _minimize_pygmo(
            problem=problem,
            x0=x0,
            method="bee_colony",
            specific_options=algo_specific_options,
            population_size=population_size,
            n_cores=1,
            seed=self.seed,
            discard_start_params=self.discard_start_params,
        )
        return res


@mark.minimizer(
    name="pygmo_de",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_PYGMO_INSTALLED,
    is_global=True,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class PygmoDe(Algorithm):
    population_size: int | None = None
    seed: int | None = None
    discard_start_params: bool = False
    stopping_maxiter: PositiveInt = STOPPING_MAX_ITERATIONS_GENETIC
    # TODO: Refine type to fix range [0,2]
    weight_coefficient: NonNegativeFloat = 0.8
    # TODO: Probably refine type to fix range [0,1]
    crossover_probability: NonNegativeFloat = 0.9
    mutation_variant: Literal[
        "best/1/exp",
        "rand/1/exp",
        "rand-to-best/1/exp",
        "best/2/exp",
        "rand/2/exp",
        "best/1/bin",
        "rand/1/bin",
        "rand-to-best/1/bin",
        "best/2/bin",
        "rand/2/bin",
    ] = "rand/1/exp"
    convergence_criterion_tolerance: NonNegativeFloat = 1e-6
    convergence_relative_params_tolerance: NonNegativeFloat = CONVERGENCE_XTOL_REL

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        population_size = get_population_size(
            population_size=self.population_size, x=x0, lower_bound=10
        )
        # support both integer and string specification of the mutation variant
        mutation_variant_str_to_int = {
            "best/1/exp": 1,
            "rand/1/exp": 2,
            "rand-to-best/1/exp": 3,
            "best/2/exp": 4,
            "rand/2/exp": 5,
            "best/1/bin": 6,
            "rand/1/bin": 7,
            "rand-to-best/1/bin": 8,
            "best/2/bin": 9,
            "rand/2/bin": 10,
        }
        mutation_variant = _convert_str_to_int(
            str_to_int=mutation_variant_str_to_int, value=self.mutation_variant
        )

        algo_specific_options = {
            "gen": self.stopping_maxiter,
            "F": self.weight_coefficient,
            "CR": self.crossover_probability,
            "variant": mutation_variant,
            "ftol": self.convergence_criterion_tolerance,
            "xtol": self.convergence_relative_params_tolerance,
        }

        res = _minimize_pygmo(
            problem=problem,
            x0=x0,
            method="de",
            specific_options=algo_specific_options,
            population_size=population_size,
            n_cores=1,
            seed=self.seed,
            discard_start_params=self.discard_start_params,
        )
        return res


@mark.minimizer(
    name="pygmo_sea",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_PYGMO_INSTALLED,
    is_global=True,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class PygmoSea(Algorithm):
    population_size: int | None = None
    seed: int | None = None
    discard_start_params: bool = False
    stopping_maxiter: PositiveInt = (
        10_000  # Each generation will compute the objective once
    )

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        population_size = get_population_size(
            population_size=self.population_size, x=x0, lower_bound=10
        )

        algo_specific_options = {
            "gen": self.stopping_maxiter,
        }

        res = _minimize_pygmo(
            problem=problem,
            x0=x0,
            method="sea",
            specific_options=algo_specific_options,
            population_size=population_size,
            n_cores=1,
            seed=self.seed,
            discard_start_params=self.discard_start_params,
        )
        return res


@mark.minimizer(
    name="pygmo_sga",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_PYGMO_INSTALLED,
    is_global=True,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class PygmoSga(Algorithm):
    population_size: int | None = None
    seed: int | None = None
    discard_start_params: bool = False
    stopping_maxiter: PositiveInt = STOPPING_MAX_ITERATIONS_GENETIC
    # TODO: Refine type to fix range [0,1]
    crossover_probability: NonNegativeFloat = 0.9
    crossover_strategy: Literal[
        "exponential",
        "sbx",
        "single",
        "binomial",
    ] = "exponential"
    # TODO: Refine type to fix range [1,100]
    eta_c: PositiveFloat | None = None
    # TODO: Refine type to fix range [0,1]
    mutation_probability: NonNegativeFloat = 0.02
    mutation_strategy: Literal["uniform", "polynomial"] = "polynomial"
    # TODO: Refine type to fix range [0,1]
    mutation_polynomial_distribution_index: NonNegativeFloat | None = None
    # TODO: Refine type to fix range [0,1]
    mutation_gaussian_width: NonNegativeFloat | None = None
    selection_strategy: Literal["tournament", "truncated"] = "tournament"
    # TODO: Check if should be NonNegativeInt
    selection_truncated_n_best: int | None = None
    # TODO Check if should be NonNegativeInt
    selection_tournament_size: int | None = None

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        population_size = get_population_size(
            population_size=self.population_size, x=x0, lower_bound=64
        )

        if self.eta_c is not None and self.crossover_strategy != "sbx":
            warnings.warn(
                f"You specified crossover strategy {self.crossover_strategy}"
                "and eta_c. However, eta_c is ignored because it is only used when "
                "the crossover_strategy is set to sbx."
            )
        eta_c = 1.0 if self.eta_c is None else self.eta_c

        if (
            self.mutation_polynomial_distribution_index is not None
        ) and self.mutation_strategy != "polynomial":
            warnings.warn(
                "You specified a mutation_polynomial_distribution_index but"
                "did not choose polynomial as your mutation_strategy. Thus, "
                "mutation_polynomial_distribution_index will be ignored."
            )
        if (
            self.mutation_gaussian_width is not None
            and self.mutation_strategy != "gaussian"
        ):
            warnings.warn(
                "You specified a mutation_gaussian_width but "
                "did not choose gaussion as your mutation_strategy. "
                "Thus, mutation_gaussian_width will be ignored."
            )
        if (
            self.selection_strategy != "truncated"
            and self.selection_truncated_n_best is not None
        ):
            warnings.warn(
                "You specified selection_truncated_n_best but "
                "did not specify truncated as your selection strategy. "
                "Therefore, selection_truncated_n_best is ignored."
            )
        if (
            self.selection_strategy != "tournament"
            and self.selection_tournament_size is not None
        ):
            warnings.warn(
                "You specified selection_tournament_size but "
                "did not specify tournament as your selection strategy. "
                "Therefore, selection_tournament_size is ignored."
            )

        if (
            self.mutation_strategy == "gaussian"
            and self.mutation_gaussian_width is not None
        ):
            param_m = self.mutation_gaussian_width
        elif (
            self.mutation_strategy == "polynomial"
            and self.mutation_polynomial_distribution_index is not None
        ):
            param_m = self.mutation_polynomial_distribution_index
        else:
            param_m = 1.0

        if (
            self.selection_strategy == "truncated"
            and self.selection_truncated_n_best is not None
        ):
            param_s = self.selection_truncated_n_best
        elif (
            self.selection_strategy == "tournament"
            and self.selection_tournament_size is not None
        ):
            param_s = self.selection_tournament_size
        else:
            param_s = 2

        algo_specific_options = {
            "gen": self.stopping_maxiter,
            "cr": self.crossover_probability,
            "eta_c": eta_c,
            "m": self.mutation_probability,
            "param_m": param_m,
            "crossover": self.crossover_strategy,
            "mutation": self.mutation_strategy,
            "selection": self.selection_strategy,
            "param_s": param_s,
        }

        res = _minimize_pygmo(
            problem=problem,
            x0=x0,
            method="sga",
            specific_options=algo_specific_options,
            population_size=population_size,
            n_cores=1,
            seed=self.seed,
            discard_start_params=self.discard_start_params,
        )
        return res


@mark.minimizer(
    name="pygmo_sade",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_PYGMO_INSTALLED,
    is_global=True,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class PygmoSade(Algorithm):
    population_size: int | None = None
    seed: int | None = None
    discard_start_params: bool = False
    jde: bool = True
    stopping_maxiter: PositiveInt = STOPPING_MAX_ITERATIONS_GENETIC
    mutation_variant: Literal[
        "best/1/exp",
        "rand/1/exp",
        "rand-to-best/1/exp",
        "best/2/exp",
        "rand/2/exp",
        "best/1/bin",
        "rand/1/bin",
        "rand-to-best/1/bin",
        "best/2/bin",
        "rand/2/bin",
        "rand/3/exp",
        "rand/3/bin",
        "best/3/exp",
        "best/3/bin",
        "rand-to-current/2/exp",
        "rand-to-current/2/bin",
        "rand-to-best-and-current/2/exp",
        "rand-to-best-and-current/2/bin",
    ] = "rand/1/exp"
    keep_adapted_params: bool = False
    ftol: NonNegativeFloat = 1e-6
    xtol: NonNegativeFloat = 1e-6

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        population_size = get_population_size(
            population_size=self.population_size, x=x0, lower_bound=64
        )
        mutation_variant_str_to_int = {
            "best/1/exp": 1,
            "rand/1/exp": 2,
            "rand-to-best/1/exp": 3,
            "best/2/exp": 4,
            "rand/2/exp": 5,
            "best/1/bin": 6,
            "rand/1/bin": 7,
            "rand-to-best/1/bin": 8,
            "best/2/bin": 9,
            "rand/2/bin": 10,
            "rand/3/exp": 11,
            "rand/3/bin": 12,
            "best/3/exp": 13,
            "best/3/bin": 14,
            "rand-to-current/2/exp": 15,
            "rand-to-current/2/bin": 16,
            "rand-to-best-and-current/2/exp": 17,
            "rand-to-best-and-current/2/bin": 18,
        }
        mutation_variant = _convert_str_to_int(
            str_to_int=mutation_variant_str_to_int, value=self.mutation_variant
        )

        algo_specific_options = {
            "gen": self.stopping_maxiter,
            "variant": mutation_variant,
            "variant_adptv": 1 if self.jde else 2,
            "ftol": self.ftol,
            "xtol": self.xtol,
            "memory": self.keep_adapted_params,
        }

        res = _minimize_pygmo(
            problem=problem,
            x0=x0,
            method="sade",
            specific_options=algo_specific_options,
            population_size=population_size,
            n_cores=1,
            seed=self.seed,
            discard_start_params=self.discard_start_params,
        )
        return res


@mark.minimizer(
    name="pygmo_cmaes",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_PYGMO_INSTALLED,
    is_global=True,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class PygmoCmaes(Algorithm):
    population_size: int | None = None
    seed: int | None = None
    discard_start_params: bool = False
    stopping_maxiter: PositiveInt = STOPPING_MAX_ITERATIONS_GENETIC
    # TODO: Refine type to fix range [0,1]
    backward_horizon: NonNegativeFloat | None = None
    # TODO: Refine type to fix range [0,1]
    variance_loss_compensation: NonNegativeFloat | None = None
    # TODO: Refine type to fix range [0,1]
    learning_rate_rank_one_update: NonNegativeFloat | None = None
    # TODO: Refine type to fix range [0,1]
    learning_rate_rank_mu_update: NonNegativeFloat | None = None
    # TODO: Check if should be NonNegativeFloat
    initial_step_size: float = 0.5
    ftol: NonNegativeFloat = 1e-6
    xtol: NonNegativeFloat = 1e-6
    keep_adapted_params: bool = False

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        population_size = get_population_size(
            population_size=self.population_size, x=x0, lower_bound=64
        )

        algo_specific_options = {
            "gen": self.stopping_maxiter,
            "cc": self.backward_horizon if self.backward_horizon is not None else -1.0,
            "cs": self.variance_loss_compensation
            if self.variance_loss_compensation is not None
            else -1.0,
            "c1": self.learning_rate_rank_one_update
            if self.learning_rate_rank_one_update is not None
            else -1.0,
            "cmu": self.learning_rate_rank_mu_update
            if self.learning_rate_rank_mu_update is not None
            else -1.0,
            "sigma0": self.initial_step_size,
            "ftol": self.ftol,
            "xtol": self.xtol,
            "memory": self.keep_adapted_params,
            "force_bounds": True,
        }

        res = _minimize_pygmo(
            problem=problem,
            x0=x0,
            method="cmaes",
            specific_options=algo_specific_options,
            population_size=population_size,
            n_cores=1,
            seed=self.seed,
            discard_start_params=self.discard_start_params,
        )
        return res


@mark.minimizer(
    name="pygmo_simulated_annealing",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_PYGMO_INSTALLED,
    is_global=True,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class PygmoSimulatedAnnealing(Algorithm):
    population_size: int | None = None
    seed: int | None = None
    discard_start_params: bool = False
    start_temperature: PositiveFloat = 10.0
    # TODO: Check if type should be same as start_temperature
    end_temperature: float = 0.01
    # TODO: Check if type should be NonNegativeInt
    n_temp_adjustments: int = 10
    # TODO: Check if type should be NonNegativeInt
    n_range_adjustments: int = 10
    # TODO: Check if type should be NonNegativeInt
    bin_size: int = 10
    # TODO: Refine type to fix range [0,1]
    start_range: NonNegativeFloat = 1.0

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        population_size = get_population_size(
            population_size=self.population_size, x=x0, lower_bound=64
        )

        algo_specific_options = {
            "Ts": self.start_temperature,
            "Tf": self.end_temperature,
            "n_T_adj": self.n_temp_adjustments,
            "n_range_adj": self.n_range_adjustments,
            "bin_size": self.bin_size,
            "start_range": self.start_range,
        }

        res = _minimize_pygmo(
            problem=problem,
            x0=x0,
            method="simulated_annealing",
            specific_options=algo_specific_options,
            population_size=population_size,
            n_cores=1,
            seed=self.seed,
            discard_start_params=self.discard_start_params,
        )
        return res


@mark.minimizer(
    name="pygmo_pso",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_PYGMO_INSTALLED,
    is_global=True,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class PygmoPso(Algorithm):
    population_size: int | None = None
    seed: int | None = None
    discard_start_params: bool = False
    stopping_maxiter: PositiveInt = STOPPING_MAX_ITERATIONS_GENETIC
    # TODO: Refine type to fix range [0,1]
    omega: NonNegativeFloat = 0.7298
    # TODO: Refine type to fix range [0,4]
    force_of_previous_best: NonNegativeFloat = 2.05
    # TODO: Refine type to fix range [0,4]
    force_of_best_in_neighborhood: NonNegativeFloat = 2.05
    # TODO: Refine type to fix range [0,1]
    max_velocity: NonNegativeFloat = 0.5
    algo_variant: Literal[
        "canonical_inertia",
        "social_and_cog_rand",
        "all_components_rand",
        "one_rand",
        "canonical_constriction",
        "fips",
    ] = "canonical_constriction"
    neighbor_definition: Literal[
        "gbest",
        "lbest",
        "Von Neumann",
        "Adaptive random",
    ] = "lbest"
    neighbor_param: int | None = None
    keep_velocities: bool = False

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        if (
            self.neighbor_definition in ["gbest", "Von Neumann"]
            and self.neighbor_param is not None
        ):
            warnings.warn(
                "You gave a neighbor parameter but selected a neighbor_definition "
                "that ignores this parameter."
            )

        neighbor_param = 4 if self.neighbor_param is None else self.neighbor_param

        population_size = get_population_size(
            population_size=self.population_size, x=x0, lower_bound=10
        )

        neighbor_definition_str_to_int = {
            "gbest": 1,
            "lbest": 2,
            "Von Neumann": 3,
            "Adaptive random": 4,
        }
        algo_variant_str_to_int = {
            "canonical_inertia": 1,
            "social_and_cog_rand": 2,
            "all_components_rand": 3,
            "one_rand": 4,
            "canonical_constriction": 5,
            "fips": 6,
        }

        algo_specific_options = {
            "gen": self.stopping_maxiter,
            "omega": self.omega,
            "eta1": self.force_of_previous_best,
            "eta2": self.force_of_best_in_neighborhood,
            "max_vel": self.max_velocity,
            "variant": _convert_str_to_int(algo_variant_str_to_int, self.algo_variant),
            "neighb_type": _convert_str_to_int(
                neighbor_definition_str_to_int, self.neighbor_definition
            ),
            "neighb_param": neighbor_param,
            "memory": self.keep_velocities,
        }

        res = _minimize_pygmo(
            problem=problem,
            x0=x0,
            method="pso",
            specific_options=algo_specific_options,
            population_size=population_size,
            n_cores=1,
            seed=self.seed,
            discard_start_params=self.discard_start_params,
        )

        return res


@mark.minimizer(
    name="pygmo_pso_gen",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_PYGMO_INSTALLED,
    is_global=True,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=True,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class PygmoPsoGen(Algorithm):
    population_size: int | None = None
    n_cores: PositiveInt = 1
    seed: int | None = None
    discard_start_params: bool = False
    stopping_maxiter: PositiveInt = STOPPING_MAX_ITERATIONS_GENETIC
    # TODO: Refine type to fix range [0,1]
    omega: NonNegativeFloat = 0.7298
    # TODO: Refine type to fix range [0,4]
    force_of_previous_best: NonNegativeFloat = 2.05
    # TODO: Refine type to fix range [0,4]
    force_of_best_in_neighborhood: NonNegativeFloat = 2.05
    # TODO: Refine type to fix range [0,1]
    max_velocity: NonNegativeFloat = 0.5
    algo_variant: Literal[
        "canonical_inertia",
        "social_and_cog_rand",
        "all_components_rand",
        "one_rand",
        "canonical_constriction",
        "fips",
    ] = "canonical_constriction"
    neighbor_definition: Literal[
        "gbest",
        "lbest",
        "Von Neumann",
        "Adaptive random",
    ] = "lbest"
    neighbor_param: int | None = None
    keep_velocities: bool = False

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        if (
            self.neighbor_definition in ["gbest", "Von Neumann"]
            and self.neighbor_param is not None
        ):
            warnings.warn(
                "You gave a neighbor parameter but selected a neighbor_definition "
                "that ignores this parameter."
            )
        neighbor_param = 4 if self.neighbor_param is None else self.neighbor_param
        neighbor_str_to_int = {
            "gbest": 1,
            "lbest": 2,
            "Von Neumann": 3,
            "Adaptive random": 4,
        }
        neighbor_type = _convert_str_to_int(
            neighbor_str_to_int, self.neighbor_definition
        )
        algo_variant_str_to_int = {
            "canonical_inertia": 1,
            "social_and_cog_rand": 2,
            "all_components_rand": 3,
            "one_rand": 4,
            "canonical_constriction": 5,
            "fips": 6,
        }
        algo_variant = _convert_str_to_int(algo_variant_str_to_int, self.algo_variant)

        population_size = get_population_size(
            population_size=self.population_size, x=x0, lower_bound=10
        )

        algo_specific_options = {
            "gen": self.stopping_maxiter,
            "omega": self.omega,
            "eta1": self.force_of_previous_best,
            "eta2": self.force_of_best_in_neighborhood,
            "max_vel": self.max_velocity,
            "variant": algo_variant,
            "neighb_type": neighbor_type,
            "neighb_param": neighbor_param,
            "memory": self.keep_velocities,
        }

        res = _minimize_pygmo(
            problem=problem,
            x0=x0,
            method="pso_gen",
            specific_options=algo_specific_options,
            population_size=population_size,
            n_cores=self.n_cores,
            seed=self.seed,
            discard_start_params=self.discard_start_params,
        )
        return res


@mark.minimizer(
    name="pygmo_mbh",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_PYGMO_INSTALLED,
    is_global=True,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class PygmoMbh(Algorithm):
    population_size: int | None = None
    seed: int | None = None
    discard_start_params: bool = False
    inner_algorithm: pg.algorithm | None = None
    # this is 30 instead of 5 in pygmo for our sum of squares test to pass
    stopping_max_inner_runs_without_improvement: PositiveInt = 30
    perturbation: float = 0.01

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        # the min default population size is this large to pass our sum of
        # squares tests.
        population_size = get_population_size(
            population_size=self.population_size, x=x0, lower_bound=250
        )

        algo_specific_options = {
            "algo": self.inner_algorithm,
            "stop": self.stopping_max_inner_runs_without_improvement,
            "perturb": self.perturbation,
        }

        res = _minimize_pygmo(
            problem=problem,
            x0=x0,
            method="mbh",
            specific_options=algo_specific_options,
            population_size=population_size,
            seed=self.seed,
            discard_start_params=self.discard_start_params,
            n_cores=1,
        )
        return res


@mark.minimizer(
    name="pygmo_xnes",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_PYGMO_INSTALLED,
    is_global=True,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class PygmoXnes(Algorithm):
    population_size: float | None = None
    seed: int | None = None
    discard_start_params: bool = False
    stopping_maxiter: PositiveInt = STOPPING_MAX_ITERATIONS_GENETIC
    # TODO: Refine type to fix range [0,1]
    learning_rate_mean_update: NonNegativeFloat | None = 1.0
    # TODO: Refine type to fix range [0,1]
    learning_rate_step_size_update: NonNegativeFloat | None = None
    # TODO: Refine type to fix range [0,1]
    learning_rate_cov_matrix_update: NonNegativeFloat | None = None
    # TODO: Refine type to fix range [0,1]
    initial_search_share: NonNegativeFloat | None = 1.0
    ftol: NonNegativeFloat = 1e-6
    xtol: NonNegativeFloat = 1e-6
    keep_adapted_params: bool = False

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        population_size = get_population_size(
            population_size=self.population_size, x=x0, lower_bound=64
        )
        eta_mu = (
            -1
            if self.learning_rate_mean_update is None
            else self.learning_rate_mean_update
        )
        eta_sigma = (
            -1
            if self.learning_rate_step_size_update is None
            else self.learning_rate_step_size_update
        )
        eta_b = (
            -1
            if self.learning_rate_cov_matrix_update is None
            else self.learning_rate_cov_matrix_update
        )
        algo_specific_options = {
            "gen": self.stopping_maxiter,
            "eta_mu": eta_mu,
            "eta_sigma": eta_sigma,
            "eta_b": eta_b,
            "sigma0": self.initial_search_share,
            "ftol": self.ftol,
            "xtol": self.xtol,
            "memory": self.keep_adapted_params,
            "force_bounds": True,
        }

        res = _minimize_pygmo(
            problem=problem,
            x0=x0,
            method="xnes",
            specific_options=algo_specific_options,
            population_size=population_size,
            n_cores=1,
            seed=self.seed,
            discard_start_params=self.discard_start_params,
        )
        return res


@mark.minimizer(
    name="pygmo_gwo",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_PYGMO_INSTALLED,
    is_global=True,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class PygmoGwo(Algorithm):
    population_size: int | None = None
    seed: int | None = None
    discard_start_params: bool = False
    stopping_maxiter: PositiveInt = STOPPING_MAX_ITERATIONS_GENETIC

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        population_size = get_population_size(
            population_size=self.population_size, x=x0, lower_bound=64
        )
        algo_specific_options = {
            "gen": self.stopping_maxiter,
        }
        res = _minimize_pygmo(
            problem=problem,
            x0=x0,
            method="gwo",
            specific_options=algo_specific_options,
            population_size=population_size,
            n_cores=1,
            seed=self.seed,
            discard_start_params=self.discard_start_params,
        )
        return res


@mark.minimizer(
    name="pygmo_compass_search",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_PYGMO_INSTALLED,
    is_global=True,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class PygmoCompassSearch(Algorithm):
    population_size: int | None = None
    seed: int | None = None
    discard_start_params: bool = False
    stopping_maxfun: PositiveInt = STOPPING_MAXFUN_GLOBAL
    # TODO: Refine type to fix range (0,1]
    start_range: PositiveFloat = 0.1
    # TODO?: mus be in (0,start_range]
    stop_range: PositiveFloat = 0.01
    # TODO: Refine type to fix range (0,1)
    reduction_coeff: PositiveFloat = 0.5

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        if self.population_size is not None:
            warnings.warn(
                f"You specified population size {self.population_size}. "
                "compass_search does not have a population so this argument is ignored."
            )
            population_size = self.population_size
        else:
            # if discard_start_params is False population_size - 1
            # must still be positive
            population_size = 100

        algo_specific_options = {
            "max_fevals": self.stopping_maxfun,
            "start_range": self.start_range,
            "stop_range": self.stop_range,
            "reduction_coeff": self.reduction_coeff,
        }

        res = _minimize_pygmo(
            problem=problem,
            x0=x0,
            method="compass_search",
            specific_options=algo_specific_options,
            population_size=population_size,
            n_cores=1,
            seed=self.seed,
            discard_start_params=self.discard_start_params,
        )
        return res


@mark.minimizer(
    name="pygmo_ihs",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_PYGMO_INSTALLED,
    is_global=True,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class PygmoIhs(Algorithm):
    population_size: int | None = None
    seed: int | None = None
    discard_start_params: bool = False
    stopping_maxiter: PositiveInt = STOPPING_MAX_ITERATIONS_GENETIC
    # TODO: Probably refine type to fix range [0,1]
    choose_from_memory_probability: NonNegativeFloat = 0.85
    # TODO: Refine type to fix range [0,1]
    min_pitch_adjustment_rate: NonNegativeFloat = 0.35
    # TODO: Refine type to fix range [0,1]
    max_pitch_adjustment_rate: NonNegativeFloat = 0.99
    min_distance_bandwidth: PositiveFloat = 1e-5
    max_distance_bandwidth: PositiveFloat = 1.0

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        if self.population_size is not None:
            warnings.warn("The population size has no effect on IHS' performance.")

        population_size = get_population_size(
            population_size=self.population_size, x=x0, lower_bound=1
        )

        algo_specific_options = {
            "gen": self.stopping_maxiter,
            "phmcr": self.choose_from_memory_probability,
            "ppar_min": self.min_pitch_adjustment_rate,
            "ppar_max": self.max_pitch_adjustment_rate,
            "bw_min": self.min_distance_bandwidth,
            "bw_max": self.max_distance_bandwidth,
        }
        res = _minimize_pygmo(
            problem=problem,
            x0=x0,
            method="ihs",
            specific_options=algo_specific_options,
            population_size=population_size,
            n_cores=1,
            seed=self.seed,
            discard_start_params=self.discard_start_params,
        )
        return res


@mark.minimizer(
    name="pygmo_de1220",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_PYGMO_INSTALLED,
    is_global=True,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class PygmoDe1220(Algorithm):
    population_size: int | None = None
    seed: int | None = None
    discard_start_params: bool = False
    jde: bool = True
    stopping_maxiter: PositiveInt = STOPPING_MAX_ITERATIONS_GENETIC
    allowed_variants: List[str] | None = None
    keep_adapted_params: bool = False
    ftol: NonNegativeFloat = 1e-6
    xtol: NonNegativeFloat = 1e-6

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        variant_str_to_int = {
            "best/1/exp": 1,
            "rand/1/exp": 2,
            "rand-to-best/1/exp": 3,
            "best/2/exp": 4,
            "rand/2/exp": 5,
            "best/1/bin": 6,
            "rand/1/bin": 7,
            "rand-to-best/1/bin": 8,
            "best/2/bin": 9,
            "rand/2/bin": 10,
            "rand/3/exp": 11,
            "rand/3/bin": 12,
            "best/3/exp": 13,
            "best/3/bin": 14,
            "rand-to-current/2/exp": 15,
            "rand-to-current/2/bin": 16,
            "rand-to-best-and-current/2/exp": 17,
            "rand-to-best-and-current/2/bin": 18,
        }
        if self.allowed_variants is None:
            allowed_variant_codes = [2, 3, 7, 10, 13, 14, 15, 16]
        else:
            allowed_variant_codes = [
                _convert_str_to_int(variant_str_to_int, variant)
                for variant in self.allowed_variants
            ]

        population_size = get_population_size(
            population_size=self.population_size, x=x0, lower_bound=64
        )

        algo_specific_options = {
            "gen": self.stopping_maxiter,
            "variant_adptv": 1 if self.jde else 2,
            "ftol": self.ftol,
            "xtol": self.xtol,
            "memory": self.keep_adapted_params,
            "allowed_variants": allowed_variant_codes,
        }

        res = _minimize_pygmo(
            problem=problem,
            x0=x0,
            method="de1220",
            specific_options=algo_specific_options,
            population_size=population_size,
            n_cores=1,
            seed=self.seed,
            discard_start_params=self.discard_start_params,
        )
        return res


# ====================================================================================


def _minimize_pygmo(
    problem: InternalOptimizationProblem,
    x0: NDArray[np.float64],
    method: str,
    specific_options: dict[str, Any],
    population_size: PositiveInt,
    n_cores: int,
    seed: int | None,
    discard_start_params: bool,
) -> InternalOptimizeResult:
    if not IS_PYGMO_INSTALLED:
        raise NotInstalledError(
            f"The {method} algorithm requires the pygmo package to be installed. "
            "You can install it with 'conda install -c conda-forge pygmo'. Visit "
            "https://esa.github.io/pygmo2/install.html for more detailed installation "
            "instructions."
        )

    bounds = problem.bounds
    if bounds is None or bounds.lower is None or bounds.upper is None:
        raise ValueError(f"{method} requires finitel bounds for all parameters.")
    elif not np.isfinite(bounds.lower).all() or not np.isfinite(bounds.upper).all():
        raise ValueError(f"{method} requires finite bounds for all parameters.")

    pygmo_problem = _create_pygmo_problem(problem, len(x0), n_cores)
    algo = _create_algorithm(method, specific_options, n_cores)
    pop = _create_population(
        problem=pygmo_problem,
        population_size=population_size,
        x=x0,
        seed=seed,
        discard_start_params=discard_start_params,
    )
    evolved = algo.evolve(pop)
    result = _process_pygmo_result(evolved)
    return result


def _create_pygmo_problem(
    problem: InternalOptimizationProblem, dim: int, n_cores: int
) -> pg.problem:
    class Problem:
        def fitness(self, x):
            return [problem.fun(x)]

        def get_bounds(self):
            return (problem.bounds.lower, problem.bounds.upper)

        def gradient(self, dv):  # noqa: ARG002
            raise ValueError("No pygmo optimizer should use a gradient.")

        def batch_fitness(self, dvs):
            x_list = list(dvs.reshape(-1, dim))
            eval_list = problem.batch_fun(x_list, n_cores=n_cores)
            evals = np.array(eval_list)
            return evals

    pygmo_problem = pg.problem(Problem())
    return pygmo_problem


def _create_algorithm(
    method: str, algo_options: dict[str, Any], n_cores: int
) -> pg.algorithm:
    """Create a pygmo algorithm."""
    pygmo_uda = getattr(pg, method)
    algo = pygmo_uda(**algo_options)
    try:
        algo.set_bfe(pg.bfe())
    except AttributeError:
        if n_cores >= 2:
            warnings.warn(
                f"Your specified algorithm {method} does not support parallelization. "
                "Choose another algorithm such as pygmo_gaco to parallelize."
            )
    out = pg.algorithm(algo)
    return out


def _create_population(
    problem: InternalOptimizationProblem,
    population_size: int,
    x: NDArray[np.float64],
    seed: int | None,
    discard_start_params: bool,
) -> pg.population:
    if not discard_start_params:
        population_size = population_size - 1

    pop = pg.population(
        problem,
        size=population_size,
        seed=seed,
        b=pg.bfe(),
    )
    if not discard_start_params:
        pop.push_back(x)
    return pop


def _process_pygmo_result(evolved: pg.population) -> InternalOptimizeResult:
    result = InternalOptimizeResult(
        x=evolved.champion_x,
        fun=evolved.champion_f[0],
        success=True,
        message="Number of generations reached.",
        n_fun_evals=evolved.problem.get_fevals(),
        n_jac_evals=evolved.problem.get_gevals(),
    )

    return result


def _convert_str_to_int(str_to_int, value):
    if value in str_to_int:
        out = str_to_int[value]
    elif value not in str_to_int.values():
        raise ValueError(
            f"You specified {value} as value. "
            f"It must be one of {', '.join(str_to_int.keys())}"
        )
    else:
        out = value
    return out



from dataclasses import dataclass
from typing import Any, Callable, Literal

import numpy as np
from numpy.typing import NDArray

from optimagic import mark
from optimagic.config import IS_TRANQUILO_INSTALLED
from optimagic.exceptions import NotInstalledError
from optimagic.optimization.algorithm import Algorithm, InternalOptimizeResult
from optimagic.optimization.internal_optimization_problem import (
    InternalOptimizationProblem,
)
from optimagic.typing import (
    AggregationLevel,
    NonNegativeFloat,
    NonNegativeInt,
    PositiveFloat,
    PositiveInt,
)

if IS_TRANQUILO_INSTALLED:
    from tranquilo.options import (
        AcceptanceOptions,
        FilterOptions,
        FitterOptions,
        NoiseAdaptationOptions,
        RadiusOptions,
        SamplerOptions,
        StagnationOptions,
        SubsolverOptions,
        VarianceEstimatorOptions,
    )
    from tranquilo.tranquilo import _tranquilo
else:
    AcceptanceOptions = Any
    FilterOptions = Any
    FitterOptions = Any
    NoiseAdaptationOptions = Any
    RadiusOptions = Any
    SamplerOptions = Any
    StagnationOptions = Any
    SubsolverOptions = Any
    VarianceEstimatorOptions = Any


@mark.minimizer(
    name="tranquilo",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_TRANQUILO_INSTALLED,
    is_global=False,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=True,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=True,
)
@dataclass(frozen=True)
class Tranquilo(Algorithm):
    # function type
    functype: Literal["scalar"] = "scalar"
    # basic options
    noisy: bool = False
    # convergence options
    disable_convergence: bool = False
    convergence_ftol_abs: NonNegativeFloat = 0.0
    convergence_gtol_abs: NonNegativeFloat = 0.0
    convergence_xtol_abs: NonNegativeFloat = 0.0
    convergence_ftol_rel: NonNegativeFloat = 2e-9
    convergence_gtol_rel: NonNegativeFloat = 1e-8
    convergence_xtol_rel: NonNegativeFloat = 1e-8
    convergence_min_trust_region_radius: NonNegativeFloat = 0.0
    # stopping options
    stopping_maxfun: PositiveInt = 2_000
    stopping_maxiter: PositiveInt = 200
    stopping_maxtime: NonNegativeFloat = np.inf
    # single advanced options
    batch_evaluator: Literal[
        "joblib",
        "pathos",
    ] = "joblib"
    n_cores: PositiveInt = 1
    batch_size: PositiveInt | None = None
    sample_size: PositiveInt | None = None
    model_type: (
        Literal[
            "quadratic",
            "linear",
        ]
        | None
    ) = None
    search_radius_factor: PositiveFloat | None = None
    n_evals_per_point: NonNegativeInt | None = None
    n_evals_at_start: NonNegativeInt | None = None
    seed: int | None = 925408
    # bundled advanced options
    radius_options: RadiusOptions | None = None
    stagnation_options: StagnationOptions | None = None
    noise_adaptation_options: NoiseAdaptationOptions | None = None
    # component names and related options
    sampler: (
        Literal[
            "optimal_hull",
            "random_hull",
            "random_interior",
        ]
        | Callable
    ) = "optimal_hull"
    sampler_options: SamplerOptions | None = None
    sample_filter: (
        Literal[
            "discard_all",
            "keep_all",
            "clustering",
            "drop_excess",
        ]
        | Callable
        | None
    ) = None
    sample_filter_options: FilterOptions | None = None
    model_fitter: (
        Literal[
            "ols",
            "ridge",
            "powell",
            "tranquilo",
        ]
        | Callable
        | None
    ) = None
    model_fitter_options: FitterOptions | None = None
    cube_subsolver: (
        Literal[
            "bntr",
            "bntr_fast",
            "fallback_cube",
            "fallback_multistart",
        ]
        | Callable
    ) = "bntr_fast"
    sphere_subsolver: (
        Literal[
            "gqtpar",
            "gqtpar_fast",
            "fallback_reparametrized",
            "fallback_inscribed_cube",
            "fallback_norm_constraint",
        ]
        | Callable
    ) = "gqtpar_fast"
    retry_subproblem_with_fallback: bool = True
    subsolver_options: SubsolverOptions | None = None
    acceptance_decider: (
        Literal[
            "classic",
            "naive_noisy",
            "classic_line_search",
            "noisy",
        ]
        | Callable
        | None
    ) = None
    acceptance_decider_options: AcceptanceOptions | None = None
    variance_estimator: Literal["classic"] | Callable = "classic"
    variance_estimator_options: VarianceEstimatorOptions | None = None
    infinity_handler: Literal["relative"] | Callable = "relative"
    residualize: bool | None = None

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        if not IS_TRANQUILO_INSTALLED:
            raise NotInstalledError(
                "The 'tranquilo-ls' algorithm requires the tranquilo package "
                "to be installed. You can install it with "
                "'conda install -c conda-forge tranquilo'."
            )
        raw_res = _tranquilo(
            functype="scalar",
            criterion=problem.fun,
            x=x0,
            lower_bounds=problem.bounds.lower,
            upper_bounds=problem.bounds.upper,
            noisy=self.noisy,
            disable_convergence=self.disable_convergence,
            convergence_absolute_criterion_tolerance=self.convergence_ftol_abs,
            convergence_absolute_gradient_tolerance=self.convergence_gtol_abs,
            convergence_absolute_params_tolerance=self.convergence_xtol_abs,
            convergence_relative_criterion_tolerance=self.convergence_ftol_rel,
            convergence_relative_gradient_tolerance=self.convergence_gtol_rel,
            convergence_relative_params_tolerance=self.convergence_xtol_rel,
            convergence_min_trust_region_radius=self.convergence_min_trust_region_radius,
            stopping_max_criterion_evaluations=self.stopping_maxfun,
            stopping_max_iterations=self.stopping_maxiter,
            stopping_max_time=self.stopping_maxtime,
            batch_evaluator=self.batch_evaluator,
            n_cores=self.n_cores,
            batch_size=self.batch_size,
            sample_size=self.sample_size,
            model_type=self.model_type,
            search_radius_factor=self.search_radius_factor,
            n_evals_per_point=self.n_evals_per_point,
            n_evals_at_start=self.n_evals_at_start,
            seed=self.seed,
            radius_options=self.radius_options,
            stagnation_options=self.stagnation_options,
            noise_adaptation_options=self.noise_adaptation_options,
            sampler=self.sampler,
            sampler_options=self.sampler_options,
            sample_filter=self.sample_filter,
            sample_filter_options=self.sample_filter_options,
            model_fitter=self.model_fitter,
            model_fitter_options=self.model_fitter_options,
            cube_subsolver=self.cube_subsolver,
            sphere_subsolver=self.sphere_subsolver,
            retry_subproblem_with_fallback=self.retry_subproblem_with_fallback,
            subsolver_options=self.subsolver_options,
            acceptance_decider=self.acceptance_decider,
            acceptance_decider_options=self.acceptance_decider_options,
            variance_estimator=self.variance_estimator,
            variance_estimator_options=self.variance_estimator_options,
            infinity_handler=self.infinity_handler,
            residualize=self.residualize,
        )

        res = InternalOptimizeResult(
            x=raw_res["solution_x"],
            fun=raw_res["solution_criterion"],
            message=raw_res["message"],
            info={"states": raw_res["states"]},
        )
        return res


@mark.minimizer(
    name="tranquilo_ls",
    solver_type=AggregationLevel.LEAST_SQUARES,
    is_available=IS_TRANQUILO_INSTALLED,
    is_global=False,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=True,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=True,
)
@dataclass(frozen=True)
class TranquiloLS(Algorithm):
    # basic options
    noisy: bool = False
    # convergence options
    disable_convergence: bool = False
    convergence_ftol_abs: NonNegativeFloat = 0.0
    convergence_gtol_abs: NonNegativeFloat = 0.0
    convergence_xtol_abs: NonNegativeFloat = 0.0
    convergence_ftol_rel: NonNegativeFloat = 2e-9
    convergence_gtol_rel: NonNegativeFloat = 1e-8
    convergence_xtol_rel: NonNegativeFloat = 1e-8
    convergence_min_trust_region_radius: NonNegativeFloat = 0.0
    # stopping options
    stopping_maxfun: PositiveInt = 2_000
    stopping_maxiter: PositiveInt = 200
    stopping_maxtime: NonNegativeFloat = np.inf
    # single advanced options
    batch_evaluator: Literal[
        "joblib",
        "pathos",
    ] = "joblib"
    n_cores: PositiveInt = 1
    batch_size: PositiveInt | None = None
    sample_size: PositiveInt | None = None
    model_type: (
        Literal[
            "quadratic",
            "linear",
        ]
        | None
    ) = None
    search_radius_factor: PositiveFloat | None = None
    n_evals_per_point: NonNegativeInt | None = None
    n_evals_at_start: NonNegativeInt | None = None
    seed: int | None = 925408
    # bundled advanced options
    radius_options: RadiusOptions | None = None
    stagnation_options: StagnationOptions | None = None
    noise_adaptation_options: NoiseAdaptationOptions | None = None
    # component names and related options
    sampler: (
        Literal[
            "optimal_hull",
            "random_hull",
            "random_interior",
        ]
        | Callable
    ) = "optimal_hull"
    sampler_options: SamplerOptions | None = None
    sample_filter: (
        Literal[
            "discard_all",
            "keep_all",
            "clustering",
            "drop_excess",
        ]
        | Callable
        | None
    ) = None
    sample_filter_options: FilterOptions | None = None
    model_fitter: (
        Literal[
            "ols",
            "ridge",
            "powell",
            "tranquilo",
        ]
        | Callable
        | None
    ) = None
    model_fitter_options: FitterOptions | None = None
    cube_subsolver: (
        Literal[
            "bntr",
            "bntr_fast",
            "fallback_cube",
            "fallback_multistart",
        ]
        | Callable
    ) = "bntr_fast"
    sphere_subsolver: (
        Literal[
            "gqtpar",
            "gqtpar_fast",
            "fallback_reparametrized",
            "fallback_inscribed_cube",
            "fallback_norm_constraint",
        ]
        | Callable
    ) = "gqtpar_fast"
    retry_subproblem_with_fallback: bool = True
    subsolver_options: SubsolverOptions | None = None
    acceptance_decider: (
        Literal[
            "classic",
            "naive_noisy",
            "classic_line_search",
            "noisy",
        ]
        | Callable
        | None
    ) = None
    acceptance_decider_options: AcceptanceOptions | None = None
    variance_estimator: Literal["classic"] | Callable = "classic"
    variance_estimator_options: VarianceEstimatorOptions | None = None
    infinity_handler: Literal["relative"] | Callable = "relative"
    residualize: bool | None = None

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        if not IS_TRANQUILO_INSTALLED:
            raise NotInstalledError(
                "The 'tranquilo-ls' algorithm requires the tranquilo package "
                "to be installed. You can install it with "
                "'conda install -c conda-forge tranquilo'."
            )
        raw_res = _tranquilo(
            functype="least_squares",
            criterion=problem.fun,
            x=x0,
            lower_bounds=problem.bounds.lower,
            upper_bounds=problem.bounds.upper,
            noisy=self.noisy,
            disable_convergence=self.disable_convergence,
            convergence_absolute_criterion_tolerance=self.convergence_ftol_abs,
            convergence_absolute_gradient_tolerance=self.convergence_gtol_abs,
            convergence_absolute_params_tolerance=self.convergence_xtol_abs,
            convergence_relative_criterion_tolerance=self.convergence_ftol_rel,
            convergence_relative_gradient_tolerance=self.convergence_gtol_rel,
            convergence_relative_params_tolerance=self.convergence_xtol_rel,
            convergence_min_trust_region_radius=self.convergence_min_trust_region_radius,
            stopping_max_criterion_evaluations=self.stopping_maxfun,
            stopping_max_iterations=self.stopping_maxiter,
            stopping_max_time=self.stopping_maxtime,
            batch_evaluator=self.batch_evaluator,
            n_cores=self.n_cores,
            batch_size=self.batch_size,
            sample_size=self.sample_size,
            model_type=self.model_type,
            search_radius_factor=self.search_radius_factor,
            n_evals_per_point=self.n_evals_per_point,
            n_evals_at_start=self.n_evals_at_start,
            seed=self.seed,
            radius_options=self.radius_options,
            stagnation_options=self.stagnation_options,
            noise_adaptation_options=self.noise_adaptation_options,
            sampler=self.sampler,
            sampler_options=self.sampler_options,
            sample_filter=self.sample_filter,
            sample_filter_options=self.sample_filter_options,
            model_fitter=self.model_fitter,
            model_fitter_options=self.model_fitter_options,
            cube_subsolver=self.cube_subsolver,
            sphere_subsolver=self.sphere_subsolver,
            retry_subproblem_with_fallback=self.retry_subproblem_with_fallback,
            subsolver_options=self.subsolver_options,
            acceptance_decider=self.acceptance_decider,
            acceptance_decider_options=self.acceptance_decider_options,
            variance_estimator=self.variance_estimator,
            variance_estimator_options=self.variance_estimator_options,
            infinity_handler=self.infinity_handler,
            residualize=self.residualize,
        )
        res = InternalOptimizeResult(
            x=raw_res["solution_x"],
            fun=raw_res["solution_criterion"],
            message=raw_res["message"],
            info={"states": raw_res["states"]},
        )
        return res



"""Implement the fides optimizer."""

import logging
from dataclasses import dataclass
from typing import Callable, Literal, cast

import numpy as np
from numpy.typing import NDArray

from optimagic import mark
from optimagic.config import IS_FIDES_INSTALLED
from optimagic.exceptions import NotInstalledError
from optimagic.optimization.algo_options import (
    CONVERGENCE_FTOL_ABS,
    CONVERGENCE_FTOL_REL,
    CONVERGENCE_GTOL_ABS,
    CONVERGENCE_GTOL_REL,
    CONVERGENCE_XTOL_ABS,
    STOPPING_MAXITER,
)
from optimagic.optimization.algorithm import Algorithm, InternalOptimizeResult
from optimagic.optimization.internal_optimization_problem import (
    InternalOptimizationProblem,
)
from optimagic.typing import (
    AggregationLevel,
    NonNegativeFloat,
    PositiveFloat,
    PositiveInt,
)

if IS_FIDES_INSTALLED:
    from fides import Optimizer, hessian_approximation


@mark.minimizer(
    name="fides",
    solver_type=AggregationLevel.SCALAR,
    is_available=IS_FIDES_INSTALLED,
    is_global=False,
    needs_jac=True,
    needs_hess=False,
    supports_parallelism=False,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class Fides(Algorithm):
    hessian_update_strategy: Literal[
        "bfgs",
        "bb",
        "bg",
        "dfp",
        "sr1",
    ] = "bfgs"
    convergence_ftol_abs: NonNegativeFloat = CONVERGENCE_FTOL_ABS
    convergence_ftol_rel: NonNegativeFloat = CONVERGENCE_FTOL_REL
    convergence_xtol_abs: NonNegativeFloat = CONVERGENCE_XTOL_ABS
    convergence_gtol_abs: NonNegativeFloat = CONVERGENCE_GTOL_ABS
    convergence_gtol_rel: NonNegativeFloat = CONVERGENCE_GTOL_REL
    stopping_maxiter: PositiveInt = STOPPING_MAXITER
    stopping_max_seconds: float = np.inf
    trustregion_initial_radius: PositiveFloat = 1.0
    trustregion_stepback_strategy: Literal[
        "truncate",
        "reflect",
        "reflect_single",
        "mixed",
    ] = "truncate"
    trustregion_subspace_dimension: Literal[
        "full",
        "2D",
        "scg",
    ] = "full"
    trustregion_max_stepback_fraction: float = 0.95
    trustregion_decrease_threshold: float = 0.25
    trustregion_increase_threshold: float = 0.75
    trustregion_decrease_factor: float = 0.25
    trustregion_increase_factor: float = 2.0

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        res = fides_internal(
            fun_and_jac=cast(
                Callable[[NDArray[np.float64]], NDArray[np.float64]],
                problem.fun_and_jac,
            ),
            x=x0,
            lower_bounds=problem.bounds.lower,
            upper_bounds=problem.bounds.upper,
            hessian_update_strategy=self.hessian_update_strategy,
            convergence_ftol_abs=self.convergence_ftol_abs,
            convergence_ftol_rel=self.convergence_ftol_rel,
            convergence_xtol_abs=self.convergence_xtol_abs,
            convergence_gtol_abs=self.convergence_gtol_abs,
            convergence_gtol_rel=self.convergence_gtol_rel,
            stopping_maxiter=self.stopping_maxiter,
            stopping_max_seconds=self.stopping_max_seconds,
            trustregion_initial_radius=self.trustregion_initial_radius,
            trustregion_stepback_strategy=self.trustregion_stepback_strategy,
            trustregion_subspace_dimension=self.trustregion_subspace_dimension,
            trustregion_max_stepback_fraction=self.trustregion_max_stepback_fraction,
            trustregion_decrease_threshold=self.trustregion_decrease_threshold,
            trustregion_increase_threshold=self.trustregion_increase_threshold,
            trustregion_decrease_factor=self.trustregion_decrease_factor,
            trustregion_increase_factor=self.trustregion_increase_factor,
        )

        return res


def fides_internal(
    fun_and_jac: Callable[[NDArray[np.float64]], NDArray[np.float64]],
    x: NDArray[np.float64],
    lower_bounds: NDArray[np.float64] | None,
    upper_bounds: NDArray[np.float64] | None,
    hessian_update_strategy: Literal[
        "bfgs",
        "bb",
        "bg",
        "dfp",
        "sr1",
    ],
    convergence_ftol_abs: NonNegativeFloat,
    convergence_ftol_rel: NonNegativeFloat,
    convergence_xtol_abs: NonNegativeFloat,
    convergence_gtol_abs: NonNegativeFloat,
    convergence_gtol_rel: NonNegativeFloat,
    stopping_maxiter: PositiveInt,
    stopping_max_seconds: float,
    trustregion_initial_radius: PositiveFloat,
    trustregion_stepback_strategy: Literal[
        "truncate",
        "reflect",
        "reflect_single",
        "mixed",
    ],
    trustregion_subspace_dimension: Literal[
        "full",
        "2D",
        "scg",
    ],
    trustregion_max_stepback_fraction: float,
    trustregion_decrease_threshold: float,
    trustregion_increase_threshold: float,
    trustregion_decrease_factor: float,
    trustregion_increase_factor: float,
) -> InternalOptimizeResult:
    """Minimize a scalar function using the Fides Optimizer.

    For details see
    :ref: `fides_algorithm`.

    """
    if not IS_FIDES_INSTALLED:
        raise NotInstalledError(
            "The 'fides' algorithm requires the fides package to be installed. "
            "You can install it with `pip install fides>=0.7.4`."
        )

    fides_options = {
        "delta_init": trustregion_initial_radius,
        "eta": trustregion_increase_threshold,
        "fatol": convergence_ftol_abs,
        "frtol": convergence_ftol_rel,
        "gamma1": trustregion_decrease_factor,
        "gamma2": trustregion_increase_factor,
        "gatol": convergence_gtol_abs,
        "grtol": convergence_gtol_rel,
        "maxiter": stopping_maxiter,
        "maxtime": stopping_max_seconds,
        "mu": trustregion_decrease_threshold,
        "stepback_strategy": trustregion_stepback_strategy,
        "subspace_solver": trustregion_subspace_dimension,
        "theta_max": trustregion_max_stepback_fraction,
        "xtol": convergence_xtol_abs,
    }

    hessian_instance = _create_hessian_updater_from_user_input(hessian_update_strategy)

    opt = Optimizer(
        fun=fun_and_jac,
        lb=lower_bounds,
        ub=upper_bounds,
        verbose=logging.ERROR,
        options=fides_options,
        funargs=None,
        hessian_update=hessian_instance,
        resfun=False,
    )
    raw_res = opt.minimize(x)
    res = _process_fides_res(raw_res, opt)
    out = InternalOptimizeResult(
        x=res["solution_x"],
        fun=res["solution_criterion"],
        jac=res["solution_derivative"],
        hess=res["solution_hessian"],
        success=res["success"],
        message=res["message"],
        n_iterations=res["n_iterations"],
    )

    return out


def _process_fides_res(raw_res, opt):
    """Create an optimagic results dictionary from the Fides output.

    Args:
        raw_res (tuple): Tuple containing the Fides result
        opt (fides.Optimizer): Fides Optimizer after minimize has been called on it.

    """
    fval, x, grad, hess = raw_res
    res = {
        "solution_criterion": fval,
        "solution_x": x,
        "solution_derivative": grad,
        "solution_hessian": hess,
        "success": opt.converged,
        "n_iterations": opt.iteration,
        "message": _process_exitflag(opt.exitflag),
    }
    return res


def _process_exitflag(exitflag):
    messages = {
        "DID_NOT_RUN": "The optimizer did not run",
        "MAXITER": "Reached maximum number of allowed iterations",
        "MAXTIME": "Expected to reach maximum allowed time in next iteration",
        "NOT_FINITE": "Encountered non-finite fval/grad/hess",
        "EXCEEDED_BOUNDARY": "Exceeded specified boundaries",
        "DELTA_TOO_SMALL": "Trust Region Radius too small to proceed",
        "FTOL": "Converged according to fval difference",
        "XTOL": "Converged according to x difference",
        "GTOL": "Converged according to gradient norm",
    }

    out = messages.get(exitflag.name)

    return out


def _create_hessian_updater_from_user_input(hessian_update_strategy):
    hessians_needing_residuals = (
        hessian_approximation.FX,
        hessian_approximation.SSM,
        hessian_approximation.TSSM,
        hessian_approximation.GNSBFGS,
    )
    unsupported_hess_msg = (
        f"{hessian_update_strategy} not supported because it requires "
        "residuals. Choose one of 'BB', 'BFGS', 'BG', 'DFP' or 'SR1' or pass "
        "an instance of the fides.hessian_approximation.HessianApproximation "
        "class."
    )

    if hessian_update_strategy in ("broyden", "Broyden", "BROYDEN"):
        raise ValueError(
            "You cannot use the Broyden update strategy without specifying the "
            "interpolation parameter phi. Import the Broyden class from "
            "`fides.hessian_approximation`, create an instance of it with your "
            "desired value of phi and pass this instance instead."
        )
    elif isinstance(hessian_update_strategy, str):
        if hessian_update_strategy.lower() in ["fx", "ssm", "tssm", "gnsbfgs"]:
            raise NotImplementedError(unsupported_hess_msg)
        else:
            hessian_name = hessian_update_strategy.upper()
            hessian_class = getattr(hessian_approximation, hessian_name)
            hessian_instance = hessian_class()
    elif isinstance(
        hessian_update_strategy, hessian_approximation.HessianApproximation
    ):
        hessian_instance = hessian_update_strategy
        if isinstance(hessian_instance, hessians_needing_residuals):
            raise NotImplementedError(unsupported_hess_msg)
    else:
        raise TypeError(
            "You must provide a hessian_update_strategy that is either a string or an "
            "instance of the fides.hessian_approximation.HessianApproximation class."
        )
    return hessian_instance



"""Implement the POUNDERS algorithm."""

import warnings
from dataclasses import dataclass
from typing import Any, Literal

import numpy as np
from numpy.typing import NDArray

from optimagic import mark
from optimagic.config import DEFAULT_N_CORES
from optimagic.optimization.algorithm import Algorithm, InternalOptimizeResult
from optimagic.optimization.internal_optimization_problem import (
    InternalOptimizationProblem,
)
from optimagic.optimizers._pounders.pounders_auxiliary import (
    add_accepted_point_to_residual_model,
    add_geomtery_points_to_make_main_model_fully_linear,
    create_initial_residual_model,
    create_main_from_residual_model,
    evaluate_residual_model,
    find_affine_points,
    fit_residual_model,
    get_feature_matrices_residual_model,
    get_last_model_indices_and_check_for_repeated_model,
    solve_subproblem,
    update_main_model_with_new_accepted_x,
    update_residual_model,
    update_residual_model_with_new_accepted_x,
    update_trustregion_radius,
)
from optimagic.optimizers._pounders.pounders_history import LeastSquaresHistory
from optimagic.typing import (
    AggregationLevel,
    NonNegativeFloat,
    PositiveFloat,
    PositiveInt,
)


@mark.minimizer(
    name="pounders",
    solver_type=AggregationLevel.LEAST_SQUARES,
    is_available=True,
    is_global=False,
    needs_jac=False,
    needs_hess=False,
    supports_parallelism=True,
    supports_bounds=True,
    supports_linear_constraints=False,
    supports_nonlinear_constraints=False,
    disable_history=False,
)
@dataclass(frozen=True)
class Pounders(Algorithm):
    convergence_gtol_abs: NonNegativeFloat = 1e-8
    convergence_gtol_rel: NonNegativeFloat = 1e-8
    # TODO: Why can this a bool
    convergence_gtol_scaled: NonNegativeFloat | bool = False
    max_interpolation_points: PositiveInt | None = None
    # TODO: Why is this not higher?
    stopping_maxiter: PositiveInt = 2_000
    trustregion_initial_radius: PositiveFloat = 0.1
    trustregion_minimal_radius: PositiveFloat = 1e-6
    trustregion_maximal_radius: PositiveFloat = 1e6
    trustregion_shrinking_factor_not_successful: PositiveFloat = 0.5
    trustregion_expansion_factor_successful: PositiveFloat = 2
    theta1: PositiveFloat = 1e-5
    theta2: PositiveFloat = 1e-4
    trustregion_threshold_acceptance: NonNegativeFloat = 0
    trustregion_threshold_successful: NonNegativeFloat = 0.1
    c1: NonNegativeFloat | None = None
    c2: NonNegativeFloat = 10
    trustregion_subproblem_solver: Literal[
        "bntr",
        "gqtpar",
    ] = "bntr"
    trustregion_subsolver_options: dict[str, Any] | None = None
    n_cores: PositiveInt = DEFAULT_N_CORES

    def _solve_internal_problem(
        self, problem: InternalOptimizationProblem, x0: NDArray[np.float64]
    ) -> InternalOptimizeResult:
        if self.max_interpolation_points is None:
            max_interpolation_points = 2 * len(x0) + 1
        else:
            max_interpolation_points = self.max_interpolation_points

        if self.c1 is None:
            c1 = np.sqrt(x0.shape[0])
        else:
            c1 = self.c1

        if self.trustregion_subsolver_options is None:
            trustregion_subsolver_options = {}
        else:
            trustregion_subsolver_options = self.trustregion_subsolver_options

        default_options = {
            "conjugate_gradient_method": "trsbox",
            "maxiter": 50,
            "maxiter_gradient_descent": 5,
            "gtol_abs": 1e-8,
            "gtol_rel": 1e-8,
            "gtol_scaled": 0,
            "gtol_abs_cg": 1e-8,
            "gtol_rel_cg": 1e-6,
            "k_easy": 0.1,
            "k_hard": 0.2,
        }
        trustregion_subsolver_options = {
            **default_options,
            **trustregion_subsolver_options,
        }

        result = internal_solve_pounders(
            criterion=problem.fun,
            x0=x0,
            lower_bounds=problem.bounds.lower,
            upper_bounds=problem.bounds.upper,
            gtol_abs=self.convergence_gtol_abs,
            gtol_rel=self.convergence_gtol_rel,
            gtol_scaled=self.convergence_gtol_scaled,
            maxinterp=max_interpolation_points,
            maxiter=self.stopping_maxiter,
            delta=self.trustregion_initial_radius,
            delta_min=self.trustregion_minimal_radius,
            delta_max=self.trustregion_maximal_radius,
            gamma0=self.trustregion_shrinking_factor_not_successful,
            gamma1=self.trustregion_expansion_factor_successful,
            theta1=self.theta1,
            theta2=self.theta2,
            eta0=self.trustregion_threshold_acceptance,
            eta1=self.trustregion_threshold_successful,
            c1=c1,
            c2=self.c2,
            solver_sub=self.trustregion_subproblem_solver,
            conjugate_gradient_method_sub=trustregion_subsolver_options[
                "conjugate_gradient_method"
            ],
            maxiter_sub=trustregion_subsolver_options["maxiter"],
            maxiter_gradient_descent_sub=trustregion_subsolver_options[
                "maxiter_gradient_descent"
            ],
            gtol_abs_sub=trustregion_subsolver_options["gtol_abs"],
            gtol_rel_sub=trustregion_subsolver_options["gtol_rel"],
            gtol_scaled_sub=trustregion_subsolver_options["gtol_scaled"],
            gtol_abs_conjugate_gradient_sub=trustregion_subsolver_options[
                "gtol_abs_cg"
            ],
            gtol_rel_conjugate_gradient_sub=trustregion_subsolver_options[
                "gtol_rel_cg"
            ],
            k_easy_sub=trustregion_subsolver_options["k_easy"],
            k_hard_sub=trustregion_subsolver_options["k_hard"],
            batch_fun=problem.batch_fun,
            n_cores=self.n_cores,
        )

        return result


def internal_solve_pounders(
    criterion,
    x0,
    lower_bounds,
    upper_bounds,
    gtol_abs,
    gtol_rel,
    gtol_scaled,
    maxinterp,
    maxiter,
    delta,
    delta_min,
    delta_max,
    gamma0,
    gamma1,
    theta1,
    theta2,
    eta0,
    eta1,
    c1,
    c2,
    solver_sub,
    conjugate_gradient_method_sub,
    maxiter_sub,
    maxiter_gradient_descent_sub,
    gtol_abs_sub,
    gtol_rel_sub,
    gtol_scaled_sub,
    gtol_abs_conjugate_gradient_sub,
    gtol_rel_conjugate_gradient_sub,
    k_easy_sub,
    k_hard_sub,
    batch_fun,
    n_cores,
):
    """Find the local minimum to a non-linear least-squares problem using POUNDERS.

    Args:
        criterion (callable): Function that returns criterion.
        x0 (np.ndarray): Initial guess for the parameter vector (starting points).
        lower_bounds (np.ndarray): 1d array of shape (n,) with lower bounds
            for the parameter vector x.
        upper_bounds (np.ndarray): 1d array of shape (n,) with upper bounds
            for the parameter vector x.
        gtol_abs (float): Convergence tolerance for the absolute gradient norm.
            Stop if norm of the gradient is less than this.
        gtol_rel (float): Convergence tolerance for the relative gradient norm.
            Stop if norm of the gradient relative to the criterion value is less
            than this.
        gtol_scaled (float): Convergence tolerance for the scaled gradient norm.
            Stop if norm of the gradient divided by norm of the gradient at the
            initial parameters is less than this.
        maxinterp (int): Maximum number of interpolation points.
            Default is `2 * n + 1`, where `n` is the length of the parameter vector.
        maxiter (int): Maximum number of iterations. If reached, terminate.
        delta (float): Delta, initial trust-region radius.
        delta_min (float): Minimal trust-region radius.
        delta_max (float): Maximal trust-region radius.
        gamma0 (float): Shrinking factor of the trust-region radius in case the
            solution vector of the suproblem is not accepted, but the model is fully
            linar (i.e. "valid").
        gamma1 (float): Expansion factor of the trust-region radius in case the
            solution vector of the suproblem is accepted.
        theta1 (float): Threshold for adding the current candidate vector
            to the model. Function argument to find_affine_points().
        theta2 (float): Threshold for adding the current candidate vector
            to the model. Argument to get_interpolation_matrices_residual_model().
        eta0 (float): Threshold for accepting the solution vector of the trust-region
            subproblem as the best candidate.
        eta1 (float): Threshold for successfully accepting the solution vector of the
            trust-region subproblem as the best candidate.
        c1 (float): Treshold for accepting the norm of our current x candidate.
            Equal to sqrt(n) by default. Argument to find_affine_points() in case
            the input array *model_improving_points* is zero.
        c2 (int)): Treshold for accepting the norm of our current candidate vector.
            Equal to 10 by default. Argument to find_affine_points() in case
            the input array *model_improving_points* is not zero.
        solver_sub (str): Solver to use for the trust-region subproblem.
            Two internal solvers are supported:
            - "bntr": Bounded Newton Trust-Region (default, supports bound constraints)
            - "gqtpar": (does not support bound constraints)
        conjugate_gradient_method_sub (str): Method for computing the conjugate
            gradient step ("bntr").
            Available conjugate gradient methods are:
                - "cg"
                - "steihaug_toint"
                - "trsbox" (default)
        maxiter_sub (int): Maximum number of iterations in the trust-region subproblem.
        maxiter_gradient_descent_sub (int): Maximum number of gradient descent
            iterations to perform ("bntr").
        gtol_abs_sub (float): Convergence tolerance for the absolute gradient norm
            in the trust-region subproblem ("bntr").
        gtol_rel_sub (float): Convergence tolerance for the relative gradient norm
            in the trust-region subproblem ("bntr").
        gtol_scaled_sub (float): Convergence tolerance for the scaled gradient norm
            in the trust-region subproblem ("bntr").
        gtol_abs_conjugate_gradient_sub (float): Convergence tolerance for the
            absolute gradient norm in the conjugate gradient step of the trust-region
            subproblem if "cg" is used as ``conjugate_gradient_method_sub`` ("bntr").
        gtol_rel_conjugate_gradient_sub (float): Convergence tolerance for the
            relative gradient norm in the conjugate gradient step of the trust-region
            subproblem if "cg" is used as ``conjugate_gradient_method_sub`` ("bntr").
        k_easy_sub (float): Stopping criterion for the "easy" case in the trust-region
            subproblem ("gqtpar").
        k_hard_sub (float): Stopping criterion for the "hard" case in the trust-region
            subproblem ("gqtpar").
        batch_evaluator (str or callable): Name of a pre-implemented batch evaluator
            (currently 'joblib' and 'pathos_mp') or callable with the same interface
            as the optimagic batch_evaluators.
        n_cores (int): Number of processes used to parallelize the function
            evaluations. Default is 1.

    Returns:
        (dict) Result dictionary containing:
        - solution_x (np.ndarray): Solution vector of shape (n,).
        - solution_criterion (np.ndarray): Values of the criterion function at the
            solution vector. Shape (n_obs,).
        - history_x (np.ndarray): Entire history of x. Shape (history.get_n_fun(), n).
        - history_criterion (np.ndarray): Entire history of the criterion function
            evaluations. Shape (history.get_n_fun(), n_obs)
        - n_iterations (int): Number of iterations the algorithm ran before finding a
            solution vector or reaching maxiter.
        - success (bool): Boolean indicating whether a solution has been found
            before reaching maxiter.

    """
    history = LeastSquaresHistory()

    n = len(x0)
    model_indices = np.zeros(maxinterp, dtype=int)

    n_last_modelpoints = 0

    if lower_bounds is not None and upper_bounds is not None:
        if np.max(x0 + delta - upper_bounds) > 1e-10:
            raise ValueError("Starting points + delta > upper bounds.")

    xs = [x0]
    for i in range(n):
        x1 = x0.copy()
        x1[i] += delta
        xs.append(x1)

    residuals = batch_fun(x_list=xs, n_cores=n_cores)

    history.add_entries(xs, residuals)
    accepted_index = history.get_best_index()

    residual_model = create_initial_residual_model(
        history=history, accepted_index=accepted_index, delta=delta
    )
    main_model = create_main_from_residual_model(
        residual_model=residual_model, multiply_square_terms_with_intercepts=False
    )

    x_accepted = history.get_best_x()
    gradient_norm_initial = np.linalg.norm(main_model.linear_terms)
    gradient_norm_initial *= delta

    valid = True
    n_modelpoints = n + 1
    last_model_indices = np.zeros(maxinterp, dtype=int)

    converged = False
    convergence_reason = "Continue iterating."

    for niter in range(maxiter + 1):
        result_sub = solve_subproblem(
            x_accepted=x_accepted,
            main_model=main_model,
            lower_bounds=lower_bounds,
            upper_bounds=upper_bounds,
            delta=delta,
            solver=solver_sub,
            conjugate_gradient_method=conjugate_gradient_method_sub,
            maxiter=maxiter_sub,
            maxiter_gradient_descent=maxiter_gradient_descent_sub,
            gtol_abs=gtol_abs_sub,
            gtol_rel=gtol_rel_sub,
            gtol_scaled=gtol_scaled_sub,
            gtol_abs_conjugate_gradient=gtol_abs_conjugate_gradient_sub,
            gtol_rel_conjugate_gradient=gtol_rel_conjugate_gradient_sub,
            k_easy=k_easy_sub,
            k_hard=k_hard_sub,
        )

        x_candidate = x_accepted + result_sub["x"] * delta
        residuals_candidate = criterion(x_candidate)
        history.add_entries(x_candidate, residuals_candidate)

        predicted_reduction = history.get_critvals(
            accepted_index
        ) - history.get_critvals(-1)
        actual_reduction = -result_sub["criterion"]

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            rho = np.divide(predicted_reduction, actual_reduction)

        if (rho >= eta1) or (rho > eta0 and valid):
            residual_model = residual_model._replace(
                intercepts=history.get_residuals(index=accepted_index)
            )
            center_info = {"x": history.get_best_x(), "radius": delta}
            x_candidate = history.get_centered_xs(center_info, index=-1)

            residual_model = update_residual_model_with_new_accepted_x(
                residual_model=residual_model, x_candidate=x_candidate
            )
            main_model = update_main_model_with_new_accepted_x(
                main_model=main_model, x_candidate=x_candidate
            )
            x_accepted = history.get_best_x()
            accepted_index = history.get_best_index()

        critval_accepted = history.get_critvals(index=accepted_index)

        # The model is deemend "not valid" if it has less than n model points.
        # Otherwise, if the model has n points, it is considered "valid" or
        # "fully linear" or "just identified".
        # Note: valid is True in the first iteration
        if not valid:
            (
                model_improving_points,
                model_indices,
                n_modelpoints,
                project_x_onto_null,
            ) = find_affine_points(
                history=history,
                x_accepted=x_accepted,
                model_improving_points=np.zeros((n, n)),
                project_x_onto_null=False,
                delta=delta,
                theta1=theta1,
                c=c1,
                model_indices=model_indices,
                n_modelpoints=0,
            )

            if n_modelpoints < n:
                (
                    history,
                    model_indices,
                ) = add_geomtery_points_to_make_main_model_fully_linear(
                    history=history,
                    main_model=main_model,
                    model_improving_points=model_improving_points,
                    model_indices=model_indices,
                    x_accepted=x_accepted,
                    n_modelpoints=n_modelpoints,
                    delta=delta,
                    criterion=criterion,
                    lower_bounds=lower_bounds,
                    upper_bounds=upper_bounds,
                    batch_fun=batch_fun,
                    n_cores=n_cores,
                )
                n_modelpoints = n

        delta_old = delta
        delta = update_trustregion_radius(
            result_subproblem=result_sub,
            rho=rho,
            model_is_valid=valid,
            delta=delta,
            delta_min=delta_min,
            delta_max=delta_max,
            eta1=eta1,
            gamma0=gamma0,
            gamma1=gamma1,
        )

        (
            model_improving_points,
            model_indices,
            n_modelpoints,
            project_x_onto_null,
        ) = find_affine_points(
            history=history,
            x_accepted=x_accepted,
            model_improving_points=np.zeros((n, n)),
            project_x_onto_null=False,
            delta=delta,
            theta1=theta1,
            c=c1,
            model_indices=model_indices,
            n_modelpoints=0,
        )

        if n_modelpoints == n:
            valid = True
        else:
            valid = False
            (
                model_improving_points,
                model_indices,
                n_modelpoints,
                project_x_onto_null,
            ) = find_affine_points(
                history=history,
                x_accepted=x_accepted,
                model_improving_points=model_improving_points,
                project_x_onto_null=project_x_onto_null,
                delta=delta,
                theta1=theta1,
                c=c2,
                model_indices=model_indices,
                n_modelpoints=n_modelpoints,
            )

            if n_modelpoints < n:
                (
                    history,
                    model_indices,
                ) = add_geomtery_points_to_make_main_model_fully_linear(
                    history=history,
                    main_model=main_model,
                    model_improving_points=model_improving_points,
                    model_indices=model_indices,
                    x_accepted=x_accepted,
                    n_modelpoints=n_modelpoints,
                    delta=delta,
                    criterion=criterion,
                    lower_bounds=lower_bounds,
                    upper_bounds=upper_bounds,
                    batch_fun=batch_fun,
                    n_cores=n_cores,
                )

        model_indices = add_accepted_point_to_residual_model(
            model_indices, accepted_index, n_modelpoints
        )

        (
            x_sample_monomial_basis,
            monomial_basis,
            basis_null_space,
            lower_triangular,
            n_modelpoints,
        ) = get_feature_matrices_residual_model(
            history=history,
            x_accepted=x_accepted,
            model_indices=model_indices,
            delta=delta,
            c2=c2,
            theta2=theta2,
            n_maxinterp=maxinterp,
        )

        center_info = {"x": x_accepted, "radius": delta_old}
        centered_xs = history.get_centered_xs(
            center_info, index=model_indices[:n_modelpoints]
        )

        center_info = {"residuals": residual_model.intercepts}
        centered_residuals = history.get_centered_residuals(
            center_info, index=model_indices
        )

        y_residuals = evaluate_residual_model(
            centered_xs=centered_xs,
            centered_residuals=centered_residuals,
            residual_model=residual_model,
        )

        coefficients_residual_model = fit_residual_model(
            m_mat=x_sample_monomial_basis,
            n_mat=monomial_basis,
            z_mat=basis_null_space,
            n_z_mat=lower_triangular,
            y_residuals=y_residuals,
            n_modelpoints=n_modelpoints,
        )

        residual_model = residual_model._replace(
            intercepts=history.get_residuals(index=accepted_index)
        )
        residual_model = update_residual_model(
            residual_model=residual_model,
            coefficients_to_add=coefficients_residual_model,
            delta=delta,
            delta_old=delta_old,
        )

        main_model = create_main_from_residual_model(residual_model)

        gradient_norm = np.linalg.norm(main_model.linear_terms)
        gradient_norm *= delta

        (
            last_model_indices,
            n_last_modelpoints,
            same_model_used,
        ) = get_last_model_indices_and_check_for_repeated_model(
            model_indices=model_indices,
            last_model_indices=last_model_indices,
            n_modelpoints=n_modelpoints,
            n_last_modelpoints=n_last_modelpoints,
        )

        converged, convergence_reason = _check_for_convergence(
            gradient_norm=gradient_norm,
            gradient_norm_initial=gradient_norm_initial,
            critval=critval_accepted,
            delta=delta,
            delta_old=delta_old,
            same_model_used=same_model_used,
            converged=converged,
            reason=convergence_reason,
            niter=niter,
            gtol_abs=gtol_abs,
            gtol_rel=gtol_rel,
            gtol_scaled=gtol_scaled,
            maxiter=maxiter,
        )

        if converged:
            break

    result = InternalOptimizeResult(
        x=history.get_xs(index=accepted_index),
        fun=history.get_best_residuals(),
        n_iterations=niter,
        success=converged,
        message=convergence_reason,
    )

    return result


def _check_for_convergence(
    gradient_norm,
    gradient_norm_initial,
    critval,
    delta,
    delta_old,
    same_model_used,
    converged,
    reason,
    niter,
    *,
    gtol_abs,
    gtol_rel,
    gtol_scaled,
    maxiter,
):
    """Check for convergence."""
    if same_model_used and delta == delta_old:
        converged = True
        reason = "Identical model used in successive iterations."
    elif gradient_norm < gtol_abs:
        converged = True
        reason = "Norm of the gradient is less than absolute_gradient_tolerance."
    elif critval != 0 and abs(gradient_norm / critval) < gtol_rel:
        converged = True
        reason = (
            "Norm of the gradient relative to the criterion value is less than "
            "relative_gradient_tolerance."
        )
    elif (
        gradient_norm_initial != 0
        and gradient_norm / gradient_norm_initial < gtol_scaled
    ):
        converged = True
        reason = (
            "Norm of the gradient divided by norm of the gradient at the "
            "initial parameters is less than scaled_gradient_tolerance."
        )
    elif gradient_norm_initial != 0 and gradient_norm == 0 and gtol_scaled == 0:
        converged = True
        reason = (
            "Norm of the gradient divided by norm of the gradient at the "
            "initial parameters is less than scaled_gradient_tolerance."
        )
    elif critval <= -np.inf:
        converged = True
        reason = "Criterion value is negative infinity."
    elif niter == maxiter:
        reason = "Maximum number of iterations reached."

    return converged, reason



"""
Deprecated module:

Functions to read data from the database used for logging.

The functions in the module are meant for end users of optimagic. They do not require
any knowledge of databases.

When using them internally, make sure to supply a database to path_or_database.
Otherwise, the functions may be very slow.

"""

from __future__ import annotations

import warnings
from dataclasses import dataclass

from optimagic.logging.logger import SQLiteLogOptions, SQLiteLogReader


@dataclass
class OptimizeLogReader:
    def __new__(cls, *args, **kwargs):  # type: ignore
        warnings.warn(
            "OptimizeLogReader is deprecated and will be removed in a future "
            "version. Please use optimagic.logging.SQLiteLogReader instead.",
            FutureWarning,
        )
        sqlite_options = SQLiteLogOptions(*args, **kwargs)
        return SQLiteLogReader.from_options(sqlite_options)



from optimagic import deprecations
from optimagic.parameters.bounds import pre_process_bounds
from optimagic.parameters.conversion import get_converter


def count_free_params(
    params,
    constraints=None,
    bounds=None,
    # deprecated
    lower_bounds=None,
    upper_bounds=None,
):
    """Count the (free) parameters of an optimization problem.

    Args:
        params (pytree): The parameters.
        constraints (list): The constraints for the optimization problem. If constraints
            are provided, only the free parameters are counted.
        bounds: Lower and upper bounds on the parameters. The most general and preferred
            way to specify bounds is an `optimagic.Bounds` object that collects lower,
            upper, soft_lower and soft_upper bounds. The soft bounds are used for
            sampling based optimizers but are not enforced during optimization. Each
            bound type mirrors the structure of params. Check our how-to guide on bounds
            for examples. If params is a flat numpy array, you can also provide bounds
            via any format that is supported by scipy.optimize.minimize.

    Returns:
        int: Number of (free) parameters

    """
    bounds = deprecations.replace_and_warn_about_deprecated_bounds(
        bounds=bounds,
        lower_bounds=lower_bounds,
        upper_bounds=upper_bounds,
    )

    deprecations.throw_dict_constraints_future_warning_if_required(constraints)

    bounds = pre_process_bounds(bounds)
    constraints = deprecations.pre_process_constraints(constraints)

    _, internal_params = get_converter(
        params=params,
        constraints=constraints,
        bounds=bounds,
        func_eval=3,
        solver_type="value",
    )

    return int(internal_params.free_mask.sum())


def check_constraints(
    params,
    constraints,
    bounds=None,
    # deprecated
    lower_bounds=None,
    upper_bounds=None,
):
    """Raise an error if constraints are invalid or not satisfied in params.

    Args:
        params (pytree): The parameters.
        constraints (list): The constraints for the optimization problem.
        bounds: Lower and upper bounds on the parameters. The most general and preferred
            way to specify bounds is an `optimagic.Bounds` object that collects lower,
            upper, soft_lower and soft_upper bounds. The soft bounds are used for
            sampling based optimizers but are not enforced during optimization. Each
            bound type mirrors the structure of params. Check our how-to guide on bounds
            for examples. If params is a flat numpy array, you can also provide bounds
            via any format that is supported by scipy.optimize.minimize.

    Raises:
        InvalidParamsError: If constraints are valid but not satisfied.
        InvalidConstraintError: If constraints are invalid.

    """
    bounds = deprecations.replace_and_warn_about_deprecated_bounds(
        bounds=bounds,
        lower_bounds=lower_bounds,
        upper_bounds=upper_bounds,
    )

    deprecations.throw_dict_constraints_future_warning_if_required(constraints)

    bounds = pre_process_bounds(bounds)
    constraints = deprecations.pre_process_constraints(constraints)

    get_converter(
        params=params,
        constraints=constraints,
        bounds=bounds,
        func_eval=3,
        solver_type="value",
    )



"""Check option dictionaries for minimize, maximize."""


def check_optimization_options(options, usage, algorithm_mandatory=True):
    """Check optimize_options or maximize_options for usage in estimation functions."""
    options = {} if options is None else options

    if algorithm_mandatory:
        if not isinstance(options, dict) or "algorithm" not in options:
            raise ValueError(
                "optimize_options or maximize_options must be a dict containing at "
                "least the entry 'algorithm'"
            )
    else:
        if not isinstance(options, dict):
            raise ValueError(
                "optimize_options or maximize_options must be a dict or None."
            )

    criterion_options = {
        "criterion",
        "criterion_kwargs",
        "derivative",
        "derivative_kwargs",
    }

    invalid_criterion = criterion_options.intersection(options)
    if invalid_criterion:
        msg = (
            "Entries related to the criterion function, its derivatives or keyword "
            "arguments of those functions are not valid entries of optimize_options "
            f"or maximize_options for {usage}. Remove: {invalid_criterion}"
        )
        raise ValueError(msg)

    general_options = {"logging", "log_options", "constraints"}

    invalid_general = general_options.intersection(options)

    if invalid_general:
        msg = (
            "The following are not valid entries of optimize_options because they are "
            "not only relevant for minimization but also for inference: "
            f"{invalid_general}"
        )
        raise ValueError(msg)



from dataclasses import dataclass
from enum import Enum
from typing import Callable, Literal, TypedDict

from typing_extensions import NotRequired

from optimagic.config import DEFAULT_N_CORES
from optimagic.exceptions import InvalidNumdiffOptionsError


@dataclass(frozen=True)
class NumdiffOptions:
    """Options for numerical differentiation.

    Attributes:
        method: The method to use for numerical differentiation. Can be "central",
            "forward", or "backward".
        step_size: The step size to use for numerical differentiation. If None, the
            default step size will be used.
        scaling_factor: The scaling factor to use for numerical differentiation.
        min_steps: The minimum step size to use for numerical differentiation. If None,
            the default minimum step size will be used.
        n_cores: The number of cores to use for numerical differentiation.
        batch_evaluator: The batch evaluator to use for numerical differentiation. Can
            be "joblib" or "pathos", or a custom function.

    Raises:
        InvalidNumdiffError: If the numdiff options cannot be processed, e.g. because
            they do not have the correct type.

    """

    method: Literal[
        "central", "forward", "backward", "central_cross", "central_average"
    ] = "central"
    step_size: float | None = None
    scaling_factor: float = 1
    min_steps: float | None = None
    n_cores: int = DEFAULT_N_CORES
    batch_evaluator: Literal["joblib", "pathos"] | Callable = "joblib"  # type: ignore

    def __post_init__(self) -> None:
        _validate_attribute_types_and_values(self)


class NumdiffOptionsDict(TypedDict):
    method: NotRequired[
        Literal["central", "forward", "backward", "central_cross", "central_average"]
    ]
    step_size: NotRequired[float | None]
    scaling_factor: NotRequired[float]
    min_steps: NotRequired[float | None]
    n_cores: NotRequired[int]
    batch_evaluator: NotRequired[Literal["joblib", "pathos"] | Callable]  # type: ignore


def pre_process_numdiff_options(
    numdiff_options: NumdiffOptions | NumdiffOptionsDict | None,
) -> NumdiffOptions | None:
    """Convert all valid types of Numdiff options to optimagic.NumdiffOptions class.

    This just harmonizes multiple ways of specifying numdiff options into a single
    format. It performs runtime type checks, but it does not check whether numdiff
    options are consistent with other option choices.

    Args:
        numdiff_options: The user provided numdiff options.

    Returns:
        The numdiff options in the optimagic format.

    Raises:
        InvalidNumdiffOptionsError: If numdiff options cannot be processed, e.g. because
            they do not have the correct type.

    """
    if isinstance(numdiff_options, NumdiffOptions) or numdiff_options is None:
        pass
    else:
        try:
            numdiff_options = NumdiffOptions(**numdiff_options)
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception as e:
            if isinstance(e, InvalidNumdiffOptionsError):
                raise e
            raise InvalidNumdiffOptionsError(
                f"Invalid numdiff options of type: {type(numdiff_options)}. Numdiff "
                "options must be of type optimagic.NumdiffOptions, a dictionary with a"
                "subset of the keys {'method', 'step_size', 'scaling_factor', "
                "'min_steps', 'n_cores', 'batch_evaluator'}, or None."
            ) from e

    return numdiff_options


def _validate_attribute_types_and_values(options: NumdiffOptions) -> None:
    if options.method not in {
        "central",
        "forward",
        "backward",
        "central_cross",
        "central_average",
    }:
        raise InvalidNumdiffOptionsError(
            f"Invalid numdiff `method`: {options.method}. Numdiff `method` must be "
            "one of 'central', 'forward', 'backward', 'central_cross', or "
            "'central_average'."
        )

    if options.step_size is not None and (
        not isinstance(options.step_size, float) or options.step_size <= 0
    ):
        raise InvalidNumdiffOptionsError(
            f"Invalid numdiff `step_size`: {options.step_size}. Step size must be a "
            "float greater than 0."
        )

    if (
        not isinstance(options.scaling_factor, int | float)
        or options.scaling_factor <= 0
    ):
        raise InvalidNumdiffOptionsError(
            f"Invalid numdiff `scaling_factor`: {options.scaling_factor}. Scaling "
            "factor must be an integer or float greater than 0."
        )

    if options.min_steps is not None and (
        not isinstance(options.min_steps, float) or options.min_steps <= 0
    ):
        raise InvalidNumdiffOptionsError(
            f"Invalid numdiff `min_steps`: {options.min_steps}. Minimum step "
            "size must be a float greater than 0."
        )

    if not isinstance(options.n_cores, int) or options.n_cores <= 0:
        raise InvalidNumdiffOptionsError(
            f"Invalid numdiff `n_cores`: {options.n_cores}. Number of cores "
            "must be an integer greater than 0."
        )

    if not callable(options.batch_evaluator) and options.batch_evaluator not in {
        "joblib",
        "pathos",
    }:
        raise InvalidNumdiffOptionsError(
            f"Invalid numdiff `batch_evaluator`: {options.batch_evaluator}. Batch "
            "evaluator must be a callable or one of 'joblib', 'pathos'."
        )


class NumdiffPurpose(str, Enum):
    OPTIMIZE = "optimize"
    ESTIMATE_JACOBIAN = "estimate_jacobian"
    ESTIMATE_HESSIAN = "estimate_hessian"


def get_default_numdiff_options(
    purpose: NumdiffPurpose,
) -> NumdiffOptions:
    """Get default numerical derivatives options for a given purpose.

    Args:
        purpose: For what purpose the numdiff options are used.

    Returns:
        The numdiff options with defaults filled in.

    """
    defaults: NumdiffOptionsDict = {}

    if purpose == NumdiffPurpose.OPTIMIZE:
        defaults["method"] = "forward"

    if purpose == NumdiffPurpose.ESTIMATE_JACOBIAN:
        defaults["method"] = "central"

    if purpose == NumdiffPurpose.ESTIMATE_HESSIAN:
        defaults["method"] = "central_cross"
        defaults["scaling_factor"] = 2

    return NumdiffOptions(**defaults)
